""" slalom.dataops.pandasutils module """

import os

from slalom.dataops import io
from slalom.dataops import jobs
from slalom.dataops.logs import (
    get_logger,
    logged,
    logged_block,
    _get_printable_context,
    bytes_to_string,
)

logging = get_logger("slalom.dataops.sparkutils")

try:
    import pandas as pd
except Exception as ex:
    pd = None
    logging.warning(f"Could not load pandas library. Try 'pip install pandas'. {ex}")


def _raise_if_missing_pandas(as_warning=False, ex=None):
    if not pd:
        msg = f"Could not load pandas library. Try 'pip install pandas'. {ex or ''}"
        if as_warning:
            logging.warning(msg)
        else:
            raise RuntimeError(msg)


def read_csv_dir(csv_dir, usecols=None, dtype=None):
    _raise_if_missing_pandas()
    df_list = []
    for s3_path in io.list_s3_files(csv_dir):
        if "_SUCCESS" not in s3_path:
            if io.USE_SCRATCH_DIR:
                scratch_dir = io.get_scratch_dir()
                filename = os.path.basename(s3_path)
                csv_path = os.path.join(scratch_dir, filename)
                if os.path.exists(csv_path):
                    logging.info(
                        f"Skipping download of '{s3_path}'. File exists as: '{csv_path}' "
                        "(If you do not want to use this file, please delete "
                        "the file or unset the USE_SCRATCH_DIR environment variable.)"
                    )
                else:
                    logging.info(
                        f"Downloading S3 file '{s3_path}' to scratch dir: '{csv_path}'"
                    )
                io.download_s3_file(s3_path, csv_path)
            else:
                logging.info(f"Reading from S3 file: {s3_path}")
                csv_path = s3_path
            df = pd.read_csv(
                csv_path, index_col=None, header=0, usecols=usecols, dtype=dtype
            )
            df_list.append(df)
    logging.info(f"Concatenating datasets from: {csv_dir}")
    ret_val = pd.concat(df_list, axis=0, ignore_index=True)
    logging.info("Dataset concatenation was successful.")
    return ret_val


def get_pandas_df(source_path, usecols=None):
    if not pd:
        raise RuntimeError(
            "Could not execute get_pandas_df(): Pandas library not loaded."
        )
    if ".xlsx" in source_path.lower():
        df = read_excel_sheet(source_path, usecols=usecols)
    else:
        try:
            df = pd.read_csv(source_path, low_memory=False, usecols=usecols)
        except Exception as ex:
            if "Error tokenizing data. C error" in str(ex):
                logging.warning(
                    f"Failed read_csv() using default 'c' engine. "
                    f"Retrying with engine='python'...\n{ex}"
                )
                df = pd.read_csv(source_path, usecols=usecols, engine="python")
            else:
                raise ex
    return df


def read_excel_sheet(sheet_path, usecols=None):
    """
    Expects path in form of '/path/to/file.xlsx/#sheet name'
    S3 paths are excepted.
    """
    _raise_if_missing_pandas()
    filepath, sheetname = sheet_path.split("/#")
    df = pd.read_excel(filepath, sheetname=sheetname, usecols=usecols)
    return df


def print_pandas_mem_usage(df, df_name, print_fn=logging.info, min_col_size_mb=500):
    if not pd:
        logging.warning("Pandas support is not installed. Try 'pip install pandas'.")
        return None
    col_mem_usage = df.memory_usage(index=True, deep=True).sort_values(ascending=False)
    ttl_mem_usage = col_mem_usage.sum()
    col_mem_usage = col_mem_usage.nlargest(n=5)
    col_mem_usage = col_mem_usage[col_mem_usage > min_col_size_mb * 1024 * 1024]
    col_mem_usage = col_mem_usage.apply(bytes_to_string)
    msg = f"Dataframe '{df_name}' mem usage: {bytes_to_string(ttl_mem_usage)}"
    if col_mem_usage.size:
        col_usage_str = ", ".join(
            [
                f"{col}({'Index' if col == 'Index' else df[col].dtype}):{size}"
                for col, size in col_mem_usage.iteritems()
            ]
        )
        msg += f". Largest columns (over {min_col_size_mb}MB): {col_usage_str}"
    print_fn(msg)
    return msg



""" slalom.dataops.sparkutils module """

import datetime
import importlib.util
import time
import os
import sys

from pathlib import Path

import docker
import fire
import pyspark
from py4j.java_gateway import java_import
from pyspark import SparkContext, SparkConf
from pyspark.sql import SparkSession
from pyspark.sql.types import (
    StructType,
    StructField,
    DateType,
    TimestampType,
    Row as SparkRow,
)
from pyspark.sql.functions import (
    unix_timestamp,
    from_unixtime,
    to_date,
    input_file_name,
    lit,
)

from slalom.dataops import io
from slalom.dataops import dockerutils
from slalom.dataops import jobs
from slalom.dataops.logs import (
    get_logger,
    logged,
    logged_block,
    _get_printable_context,
    bytes_to_string,
)
from slalom.dataops import pandasutils

logging = get_logger("slalom.dataops.sparkutils")

try:
    import pandas as pd
except Exception as ex:
    pd = None
    logging.warning(f"Could not load pandas library. Try 'pip install pandas'. {ex}")

ENV_VAR_SPARK_UDF_MODULE = "SPARK_UDF_MODULE"

_SERVING_SPARK_REQUESTS = "serving spark requests"
ENABLE_SQL_JDBC = bool(os.environ.get("ENABLE_SQL_JDBC", False))
METASTORE_TYPE = os.environ.get("METASTORE_TYPE", "Derby")
METASTORE_SERVER = os.environ.get("METASTORE_SERVER", None) or "localhost"
METASTORE_DB_USER = os.environ.get("METASTORE_DB_USER", None)
METASTORE_DB_PASSWORD = os.environ.get("METASTORE_DB_PASSWORD", None)
SUPPORT_CLUSTER_BY = False

DOCKER_SPARK_IMAGE = os.environ.get("DOCKER_SPARK_IMAGE", "slalomggp/dataops:latest-dev")
CONTAINER_ENDPOINT = "spark://localhost:7077"
SPARK_DRIVER_MEMORY = "4g"
SPARK_EXECUTOR_MEMORY = "4g"
SPARK_WAREHOUSE_DIR = os.environ.get("SPARK_WAREHOUSE_DIR", "/spark_warehouse/data")
SPARK_S3_PREFIX = "s3a://"
SPARK_LOG_LEVEL = os.environ.get(
    "SPARK_LOG_LEVEL", "ERROR"
)  # ALL, DEBUG, ERROR, FATAL, INFO, WARN
HADOOP_HOME = os.environ.get("HADOOP_HOME", "/usr/local/hdp")
# SPARK_HOME = os.environ.get("SPARK_HOME", None)
# SPARK_CLASS_PATH = os.path.join(os.environ["SPARK_HOME"], "jars/*")

SPARK_EXTRA_AWS_JARS = [
    # Hadoop 2.7.7:
    os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/aws-java-sdk-1.7.4.jar"),
    os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/hadoop-aws-2.7.7.jar")
    # # Hadoop 3.1.2:
    # os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/aws-java-sdk-bundle-1.11.271.jar"),
    # os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/hadoop-aws-3.1.2.jar")
    # os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/aws-java-sdk-core-1.10.6.jar")
    # os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/aws-java-sdk-kms-1.10.6.jar")
    # os.path.join(HADOOP_HOME, "share/hadoop/tools/lib/aws-java-sdk-s3-1.10.6"),
]


def _add_derby_metastore_config(hadoop_conf):
    """ Returns a new hadoop_conf dict with added metastore params """
    derby_log = "/home/data/derby.log"
    derby_home = "/home/data/derby_home"
    derby_hive_metastore_dir = "/home/data/hive_metastore_db"
    for folder in [SPARK_WAREHOUSE_DIR, derby_hive_metastore_dir]:
        io.create_folder(derby_home)
    derby_options = (
        f"-Dderby.stream.error.file={derby_log} -Dderby.system.home={derby_home}"
    )
    hadoop_conf.update(
        {
            "derby.system.home": derby_home,
            "derby.stream.error.file": derby_log,
            "driver-java-options": derby_options,
            "spark.driver.extraJavaOptions": derby_options,
            "spark.executor.extraJavaOptions": derby_options,
            "hive.metastore.warehouse.dir": f"file://{derby_hive_metastore_dir}",
            # "javax.jdo.option.ConnectionURL": "jdbc:derby:memory:databaseName=metastore_db;create=true",
            "javax.jdo.option.ConnectionURL": "jdbc:derby:;databaseName=/home/data/metastore_db;create=true",
            "javax.jdo.option.ConnectionDriverName": "org.apache.derby.jdbc.EmbeddedDriver",
        }
    )
    return hadoop_conf


def _add_mysql_metastore_config(hadoop_conf):
    """ Returns a new hadoop_conf dict with added metastore params """
    hadoop_conf.update(
        {
            "javax.jdo.option.ConnectionURL": (
                f"jdbc:mysql://{METASTORE_SERVER}/"
                "metastore_db?createDatabaseIfNotExist=true&useSSL=false"
            ),
            "javax.jdo.option.ConnectionDriverName": "com.mysql.jdbc.Driver",
            "javax.jdo.option.ConnectionUserName": "root",
            "javax.jdo.option.ConnectionPassword": "root",
        }
    )
    if METASTORE_DB_USER:
        hadoop_conf["javax.jdo.option.ConnectionUserName"] = METASTORE_DB_USER
    if METASTORE_DB_PASSWORD:
        hadoop_conf["javax.jdo.option.ConnectionPassword"] = METASTORE_DB_PASSWORD
    return hadoop_conf


def _add_aws_creds_config(hadoop_conf):
    """ Returns a new hadoop_conf dict with added metastore params """
    hadoop_conf.update(
        {
            "fs.s3.impl": "org.apache.hadoop.fs.s3a.S3AFileSystem",
            "fs.s3a.impl": "org.apache.hadoop.fs.s3a.S3AFileSystem",
            "fs.s3a.endpoint": (
                f"s3.{os.environ.get('AWS_DEFAULT_REGION', 'us-east-2')}.amazonaws.com"
            ),
            "spark.jars": ",".join(SPARK_EXTRA_AWS_JARS),
            "com.amazonaws.services.s3.enableV4": "true",
        }
    )
    os.environ["HADOOP_OPTS"] = (
        os.environ.get("HADOOP_OPTS", "")
        + " -Djava.net.preferIPv4Stack=true -Dcom.amazonaws.services.s3.enableV4=true"
    )
    try:
        key, secret = io.load_aws_creds()
        logging.info(
            f"Successfully loaded AWS creds for access key: ****************{key[-4:]}"
        )
        # TODO: Confirm that these settings are not needed (avoid leaks to logs)
        # if key:
        #     hadoop_conf["fs.s3a.access.key"] = key
        # if secret:
        #     hadoop_conf["fs.s3a.secret.key"] = secret
    except Exception as ex:
        logging.info(f"Could not load AWS creds ({ex})")
    return hadoop_conf


def _get_hadoop_conf():
    hadoop_conf = {
        "spark.driver.memory": SPARK_DRIVER_MEMORY,
        "spark.executor.memory": SPARK_EXECUTOR_MEMORY,
        "spark.jars.packages": "io.delta:delta-core_2.11:0.4.0",
        "spark.logConf": "true",
        "spark.sql.warehouse.dir": SPARK_WAREHOUSE_DIR,
        "spark.ui.showConsoleProgress": "false",  # suppress updates e.g. 'Stage 2=====>'
        "log4j.rootCategory": SPARK_LOG_LEVEL,
        "log4j.logger.org.apache.hive.service.server": SPARK_LOG_LEVEL,
        "log4j.logger.org.apache.spark.api.python.PythonGatewayServer": SPARK_LOG_LEVEL,
    }
    # Add Thrift JDBC Server settings
    hadoop_conf.update(
        {
            "spark.sql.hive.thriftServer.singleSession": "true",
            "hive.server2.thrift.port": 10000,
            "hive.server2.http.endpoint": "cliservice",
            "log4j.logger.org.apache.spark.sql.hive.thriftserver": SPARK_LOG_LEVEL,
        }
    )
    hadoop_conf = _add_aws_creds_config(hadoop_conf)
    if METASTORE_TYPE.upper() == "MYSQL":
        hadoop_conf = _add_mysql_metastore_config(hadoop_conf)
    else:
        hadoop_conf = _add_derby_metastore_config(hadoop_conf)
    return hadoop_conf


# GLOBALS

spark = None
sc = None
thrift = None
_spark_container = None


@logged("starting Spark container '{spark_image}' with args: with_jupyter={with_jupyter}")
def _init_spark_container(spark_image=DOCKER_SPARK_IMAGE, with_jupyter=False):
    global _spark_container

    if _spark_container:
        return _spark_container
    port_map = {
        "4040": "4040",  # App Web UI
        "7077": "7077",  # Standalone master driver
        "8080": "8080",  # Standalone-mode master Web UI
        "8081": "8081",  # Standalone-mode worker Web UI
        "8888": "8888",  # Jupyter Notebook Server
        "10000": "10000",  # Thrift JDBC port for SQL queries
        "18080": "18080",  # History Server Web UI
    }
    io.load_aws_creds(update_env_vars=True)
    env = [
        "AWS_ACCESS_KEY_ID",
        "AWS_SECRET_ACCESS_KEY",
        "BATCH_ID=SparkContainerTest",
        "ENABLE_SQL_JDBC=True",
        "METASTORE_TYPE=MySQL",
    ]
    if "AWS_ACCESS_KEY_ID" in os.environ:
        env.append(f"AWS_ACCESS_KEY_ID={os.environ['AWS_ACCESS_KEY_ID']}")
    if "AWS_SECRET_ACCESS_KEY" in os.environ:
        env.append(f"AWS_SECRET_ACCESS_KEY={os.environ['AWS_SECRET_ACCESS_KEY']}")
    docker_client = docker.from_env()  # WSL1
    # docker_client = docker.DockerClient(base_url="npipe:////./pipe/docker_wsl")  # WSL2
    try:
        dockerutils.pull(spark_image)
    except Exception as ex:
        logging.warning(f"Could not pull latest Spark image '{spark_image}'. {ex}")
    try:
        old_container = docker_client.containers.get("spark_server")
        if old_container:
            with logged_block("terminating previous 'spark_server' docker container"):
                old_container.stop()
                logging.info("Waiting for cleanup of old Spark container...")
                time.sleep(2)
    except Exception as _:
        pass
    spark_image_cmd = "sparkutils start_server"
    if with_jupyter:
        spark_image_cmd = f"{spark_image_cmd} --with_jupyter"
    _spark_container = docker_client.containers.run(
        image=spark_image,
        name="spark_server",
        command=spark_image_cmd,
        detach=True,
        auto_remove=True,
        ports=port_map,
        environment=env,
        # stream=True,
    )
    logging.info(
        f"Attempting to initialize Spark docker container "
        f"(status={_spark_container.status})..."
    )
    MAX_WAIT_TIME = int(60 * 5)
    start = time.time()
    for line in _spark_container.logs(stream=True, until=int(start + MAX_WAIT_TIME)):
        logging.info(f"SPARK CONTAINER LOG: {line.decode('utf-8').rstrip()}")
        # time.sleep(0.2)
        if _SERVING_SPARK_REQUESTS in line.decode("utf-8"):
            logging.info(
                f"Spark container reported success after "
                f"{int(time.time() - start)} seconds"
            )
            break
        elif time.time() > start + MAX_WAIT_TIME:
            logging.info(f"Max timeout wait exceeded ({MAX_WAIT_TIME} seconds)")
            break
    if _spark_container.status in ["running", "created"]:
        return _spark_container
    else:
        raise RuntimeError(
            "Spark docker container exited unexpectedly "
            f"(status={_spark_container.status})."
        )


def _destroy_spark_container():
    global _spark_container

    if _spark_container:
        _spark_container.stop()
        _spark_container = None


@logged(
    "initializing Spark with args: dockerized={dockerized}, with_jupyter={with_jupyter}"
)
def _init_spark(dockerized=False, with_jupyter=False, daemon=False):
    """Return an initialized spark object"""
    global spark, sc, thrift

    if dockerized:
        container = _init_spark_container(with_jupyter=with_jupyter)
        # context = SparkContext(conf=conf)
        os.environ["PYSPARK_PYTHON"] = sys.executable
        with logged_block("connecting to spark container"):
            spark = SparkSession.builder.master(CONTAINER_ENDPOINT).getOrCreate()
        spark.sparkContext.setLogLevel(SPARK_LOG_LEVEL)
        sc = spark.sparkContext
    elif daemon:
        cmd = f"{sys.executable} -m slalom.dataops.sparkutils start_server"
        wait_test = lambda line: _SERVING_SPARK_REQUESTS in line
        wait_max = 120  # Max wait in seconds
        if with_jupyter:
            cmd = f"{cmd} --with_jupyter"
        jobs.run_command(cmd, daemon=True, wait_test=wait_test, wait_max=wait_max)
    else:
        _init_local_spark()


def _init_local_spark():
    """Return an initialized local spark object"""
    global spark, sc, thrift

    # context = SparkContext(conf=conf)
    for folder in [SPARK_WAREHOUSE_DIR]:
        io.create_folder(folder)
    conf = SparkConf()
    hadoop_conf = _get_hadoop_conf()
    for fn in [conf.set]:
        # for fn in [conf.set, SparkContext.setSystemProperty, context.setSystemProperty]:
        for k, v in hadoop_conf.items():
            fn(k, v)
    os.environ["PYSPARK_PYTHON"] = sys.executable
    with logged_block("creating spark session"):
        spark = (
            SparkSession.builder.config(conf=conf)
            .master("local")
            .appName("Python Spark")
            .enableHiveSupport()
            .getOrCreate()
        )
        sc = spark.sparkContext
        # Set the property for the driver. Doesn't work using the same syntax
        # as the executor because the jvm has already been created.
        sc.setSystemProperty("com.amazonaws.services.s3.enableV4", "true")
    if not ENABLE_SQL_JDBC:
        logging.info(f"Skipping Thrift server launch (ENABLE_SQL_JDBC={ENABLE_SQL_JDBC})")
        thrift = None
    else:
        with logged_block("starting Thrift server"):
            java_import(sc._gateway.jvm, "")
            spark_hive = sc._gateway.jvm.org.apache.spark.sql.hive
            thrift_class = spark_hive.thriftserver.HiveThriftServer2
            thrift = thrift_class.startWithContext(spark._jwrapped)
        logging.info("Sleeping while waiting for Thrift Server...")
        time.sleep(1)
    spark.sparkContext.setLogLevel(SPARK_LOG_LEVEL)
    _print_conf_debug(sc)
    if ENV_VAR_SPARK_UDF_MODULE in os.environ:
        add_udf_module(os.environ.get(ENV_VAR_SPARK_UDF_MODULE))
    else:
        logging.info("Skipping loading UDFs (env variable not set)")
    for jar_path in SPARK_EXTRA_AWS_JARS:
        sc.addPyFile(jar_path)


@logged("importing from dynamic python file '{absolute_file_path}'")
def path_import(absolute_file_path):
    """implementation taken from https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly"""
    module_name = os.path.basename(absolute_file_path)
    module_name = ".".join(module_name.split(".")[:-1])  # removes .py suffix
    spec = importlib.util.spec_from_file_location(module_name, absolute_file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


@logged("loading UDFs from module directory '{module_dir}'")
def add_udf_module(module_dir=None):
    """
    Add a package from module_dir (or zip file) and register any udfs within the package
    The module must contain a '__init__.py' file and functions to be imported should be
    annotated using the @udf() decorator.
    # https://stackoverflow.com/questions/47558704/python-dynamic-import-methods-from-file
    """
    global sc
    from inspect import getmembers, isfunction

    # module_dir = module_dir or os.environ.get(ENV_VAR_SPARK_UDF_MODULE)
    module_dir = os.path.realpath(module_dir)
    # module_root = Path(module_dir).parent
    # module_name = os.path.basename(module_dir)
    # if module_root not in sys.path:
    #     sys.path.append(module_root)
    if not os.path.isdir(module_dir):
        raise ValueError(f"Folder '{module_dir}' does not exist.")
    for file in io.list_files(module_dir):
        if file.endswith(".py"):
            module = path_import(file)
            for member in getmembers(module):
                if isfunction(member[1]):
                    logging.info(f"Found module function: {member}")
                    func_name, func = member[0], member[1]
                    if func_name[:1] != "_" and func_name != "udf":
                        logging.info(f"Registering UDF '{func_name}':\n{func.__dict__}")
                        spark.udf.register(func_name, func)
                # else:
                #     logging.info(f"Found module entity: {member}")
    # sc.addPyFile(jar_path)


@logged("starting Jupyter notebooks server")
def start_jupyter(nb_directory="/home/jovyan/work", nb_token="qwerty123"):
    jupyter_run_command = (
        f"jupyter lab"
        f" --NotebookApp.notebook_dir='{nb_directory}'"
        f" --NotebookApp.token='{nb_token}'"
        f" --allow-root"
    )
    log_file = "jupyter_log.txt"
    jobs.run_command(jupyter_run_command, daemon=True, log_file_path=log_file)
    time.sleep(5)
    logging.info("\nJUPYTER_LOG:".join(io.get_text_file_contents(log_file).splitlines()))
    logging.info(
        "Jupyter notebooks server started at: https://localhost:8888/?token=qwerty123"
    )


def get_spark(dockerized=False):
    global spark

    if not spark:
        _init_spark(dockerized=dockerized)
    return spark


def _print_conf_debug(sc):
    """ Print all spark and hadoop config settings """
    logging.debug(
        "SparkSession 'spark' and SparkContext 'sc' initialized with settings:\n"
        f"{_get_printable_context(dict(sc._conf.getAll()))}"
    )


# Spark Helper Function:
@logged("creating table '{table_name}'", success_detail="{result.count():,.0f} rows")
def create_spark_sql_table(
    table_name,
    sql,
    print_row_count=True,
    print_n_rows=None,
    run_audit=True,
    schema_only=False,
):
    spark.sql(f"DROP TABLE IF EXISTS {table_name}")
    distribution_clause = ""
    for col in ["AccountId", "OpportunityId"]:
        if SUPPORT_CLUSTER_BY and not distribution_clause and col in sql:
            distribution_clause = "\n    DISTRIBUTE BY {col}"
    sql_command = f"""
    CREATE TABLE {table_name}
    USING PARQUET
    AS
    {sql}
    {distribution_clause}
    """
    spark.sql(sql_command)
    df = spark.sql(f"SELECT * FROM {table_name}")
    if print_n_rows:
        sample_spark_table(table_name, n=print_n_rows)
    if run_audit:
        audit_spark_table_keys(table_name)
    return df


def audit_spark_table_keys(table_name, key_col_suffix="Id", raise_error=False):
    df = spark.sql(f"SELECT * FROM {table_name}")
    key_cols = [c for c in df.columns if key_col_suffix in c]
    if not key_cols:
        key_cols.append(df.columns[0])
    cols = ",".join(
        [
            f"COUNT(DISTINCT {c}) AS {c}__values,\nCOUNT(*) - COUNT({c}) as {c}__null"
            for c in key_cols
        ]
    )
    sql = f"SELECT COUNT(*) AS __num_rows, {cols}\nFROM {table_name}"
    logging.info(f"Running '{table_name}' table audit...")
    result = spark.sql(sql).collect()[0]
    num_rows = result["__num_rows"]
    unique = []
    empty = []
    for col in key_cols:
        if result[col + "__null"] >= num_rows:
            empty.append(col)
        elif result[col + "__values"] >= num_rows - 1:
            unique.append(col)
    result_text = (
        f"Found unique column(s) [{','.join(unique) or '(none)'}] "
        f"and empty columns [{','.join(empty) or '(none)'}]. "
        f"Table profile: {result}"
    )
    if not unique:
        failure_msg = f"Audit failed for table '{table_name}'. {result_text}"
        if raise_error:
            raise RuntimeError(failure_msg)
        else:
            logging.warning(f"Table audit warning for '{table_name}'. {result_text}")
    elif len(empty):
        logging.warning(f"Table audit warning for '{table_name}'. {result_text}")
    else:
        logging.info(f"Table audit successful for '{table_name}'. {result_text}")


def sample_spark_table(table_name, n=1):
    df = spark.sql(f"SELECT * FROM {table_name} LIMIT {n}")
    sample_spark_df(df, n=n, name=table_name)


def sample_spark_df(df, n=1, name=None, log_fn=logging.debug):
    log_fn(
        f"Spark Dataframe column list: "
        f"{', '.join(['{dtype[0]} ({dtype[1]})' for dtype in df.dtypes])}"
        f"'{name or 'Dataframe'}' row sample:\n{df.limit(n).toPandas().head(n)}\n"
    )


def create_spark_table(
    df, table_name, print_n_rows=None, run_audit=False, schema_only=False
):
    start_time = time.time()
    if isinstance(df, pyspark.sql.DataFrame):
        logging.info(f"Creating spark table '{table_name}' from spark dataframe...")
        spark_df = df
    elif pd and isinstance(df, pd.DataFrame):
        # Coerce all values in string columns to string
        logging.debug("Coercing column types string prior to save...")
        for col in df.select_dtypes(["object"]):
            df[col] = df[col].astype("str")
        logging.debug("Converting pandas dataframe to spark dataframe prior to save...")
        spark_df = spark.createDataFrame(df)
        logging.info(f"Creating spark table '{table_name}' from pandas dataframe...")
    else:
        logging.info(
            f"Creating table '{table_name}' from unknown type '{type(df).__name__}"
        )
        spark_df = spark.createDataFrame(df, verifySchema=False)
    spark_df.write.saveAsTable(table_name, mode="overwrite")
    if print_n_rows:
        sample_spark_table(table_name, n=print_n_rows)
    if run_audit:
        audit_spark_table_keys(table_name)


def _verify_path(file_path):
    return file_path.replace(
        "s3://", SPARK_S3_PREFIX
    )  # .replace("propensity-to-buy", "propensity-to-buy-2")


@logged("loading spark table '{table_name}'")
def load_to_spark_table(
    table_name,
    file_path,
    entity_type=None,
    infer_schema=True,
    date_format=None,
    timestamp_format=None,
    filename_column="filename",
    df_cleanup_function=None,
    print_n_rows=None,
    clean_col_names=False,
    schema_only=False,
):
    start_time = time.time()
    file_path = _verify_path(file_path)

    if ".xlsx" in file_path.lower():
        if pd:
            logging.debug(
                f"Using pandas to load spark table '{table_name}' from '{file_path}'..."
            )
            df = pandasutils.get_pandas_df(file_path)
            create_spark_table(df, table_name, print_n_rows=print_n_rows)
        else:
            pandasutils._raise_if_missing_pandas()
    else:
        logging.debug(f"Loading spark table '{table_name}' from file '{file_path}'...")
        df = spark.read.csv(
            file_path,
            header=True,
            escape='"',
            quote='"',
            multiLine=True,
            inferSchema=True,
            enforceSchema=False,
            dateFormat=date_format,
            timestampFormat=timestamp_format,
            columnNameOfCorruptRecord="__READ_ERRORS",
        )
        if filename_column:
            df = df.withColumn(filename_column, input_file_name())
        if df_cleanup_function:
            df = df_cleanup_function(df)
        create_spark_table(
            df,
            table_name,
            print_n_rows=print_n_rows,
            run_audit=False,
            schema_only=schema_only,
        )


@logged("saving '{table_name}' to file")
def save_spark_table(
    table_name,
    file_path,
    entity_type=None,
    force_single_file=False,
    compression="gzip",
    schema_only=True,
    overwrite=True,
):
    start_time = time.time()
    file_path = _verify_path(file_path)
    df = spark.sql(f"SELECT * FROM {table_name}")
    if io.file_exists(os.path.join(file_path, "_SUCCESS")):
        if overwrite:
            logging.warning(
                "Saved table already exists and overwrite=True. Deleting older files."
            )
            for oldfile in io.list_files(file_path):
                io.delete_file(oldfile)
    if force_single_file:
        logging.debug(
            f"Saving spark table '{table_name}' to single file: '{file_path}'..."
        )
        df = df.coalesce(1)
    else:
        logging.debug(f"Saving spark table '{table_name}' to folder: '{file_path}'...")
    try:
        df.write.csv(  # SAFE
            file_path,
            mode="overwrite",
            header=True,
            compression=compression,
            quote='"',
            escape='"',
        )
    except Exception as ex:  # intermittent failures can be caused by eventual consistency
        logging.warning(
            f"Retrying S3 table save operation because the first attempt failed ({ex})"
        )
        time.sleep(20)  # Sleep to allow S3 to reach eventual consistency
        df.write.csv(  # SAFE
            file_path,
            mode="overwrite",
            header=True,
            compression=compression,
            quote='"',
            escape='"',
        )


def get_spark_table_as_pandas(table_name):
    if not pd:
        raise RuntimeError(
            "Could not execute get_pandas_from_spark_table(): Pandas library not loaded."
        )
    return spark.sql(f"SELECT * FROM {table_name}").toPandas()


# Create Dates table
def create_calendar_table(table_name, start_date, end_date):
    num_days = (end_date - start_date).days
    date_rows = [
        SparkRow(start_date + datetime.timedelta(days=n)) for n in range(0, num_days)
    ]
    df = spark.createDataFrame(date_rows)
    df = df.selectExpr("_1 AS calendar_date", "date_format(_1, 'yMMdd') AS YYYYMMDD")
    create_spark_table(df, table_name)


@logged(
    "starting spark server with args:"
    " dockerized={dockerized}, with_jupyter={with_jupyter}"
)
def start_server(dockerized: bool = None, with_jupyter: bool = True, daemon: bool = None):
    if dockerized is None:
        dockerized = (
            False
            if any(["SPARK_HOME" in os.environ, "HADOOP_CONF_DIR" in os.environ])
            else True
        )
    if daemon is None:
        daemon = dockerized
    if dockerized:
        container = _init_spark_container(with_jupyter=with_jupyter)
        if daemon:
            logging.info("Now serving spark requests via docker.")
        else:
            with logged_block("hosting spark container"):
                while True:
                    time.sleep(30)
    else:
        _init_spark(dockerized=False, with_jupyter=with_jupyter, daemon=daemon)
        logging.info(
            "Spark server started. "
            "Monitor via http://localhost:4040 or http://127.0.0.1:4040"
        )
        # if with_jupyter:
        #     start_jupyter()
        # else:
        #     logging.info("Skipping Jupyter notebooks server launch...")
        if not daemon:
            with logged_block(_SERVING_SPARK_REQUESTS):
                # NOTE: When run containerized, the above message triggers
                #       the host to stop echoing logs
                while True:
                    time.sleep(30)


def main():
    fire.Fire({"start_server": start_server, "start_jupyter": start_jupyter})


if __name__ == "__main__":
    main()



#!bin/env python3

import re
import os
import json
import yaml
import sys
from pathlib import Path
from collections import OrderedDict
import fire

from slalom.dataops import env, io, jobs
from slalom.dataops.logs import logged, logged_block, get_logger

IMAGE_BASE = "slalomggp/singer"
SINGER_PLUGINS_INDEX = os.environ.get("SINGER_PLUGINS_INDEX", "./singer_index.yml")

logging = get_logger("slalom.dataops.taputils")

try:
    import docker
    from slalom.dataops import dockerutils
except Exception as ex:
    docker = None
    dockerutils = None
    logging.warning("Docker libraries were not able to be loaded ({ex}).")


# ROOT_DIR = "."
_ROOT_DIR = "/projects/my-project"
VENV_ROOT = "/venv"
INSTALL_ROOT = "/usr/bin"
BASE_DOCKER_IMAGE = "slalomggp/singer"


def _get_root_dir():
    return "."


def _get_secrets_dir():
    result = os.environ.get("TAP_SECRETS_DIR", f"{_get_root_dir()}/.secrets")
    io.create_folder(result)
    return result


def _get_scratch_dir():
    result = os.environ.get("TAP_SCRATCH_DIR", f"{_get_root_dir()}/.output")
    io.create_folder(result)
    return result


def _get_root_dir():
    return _ROOT_DIR


def _get_secrets_dir():
    return os.environ.get("TAP_SECRETS_DIR", f"{_get_root_dir()}/.secrets")


def _get_scratch_dir():
    return os.environ.get("TAP_SCRATCH_DIR", f"{_get_root_dir()}/.output")


def _get_taps_dir(override=None):
    taps_dir = override or os.environ.get("TAP_CONFIG_DIR", ".")
    return io.make_local(taps_dir)  # if remote path provided, download locally


def _get_catalog_output_dir(tap_name):
    result = f"{_get_scratch_dir()}/taps/{tap_name}-catalog"
    io.create_folder(result)
    return result


def _get_plan_file(tap_name, taps_dir=None):
    return os.path.join(_get_taps_dir(taps_dir), f"data-plan-{tap_name}.yml")


def _get_select_file(taps_dir=None):
    return os.path.join(_get_taps_dir(taps_dir), f"data.select")


def _get_config_file(plugin_name, config_dir=None):
    """
    Returns a path to the configuration file which also contains secrets.
     - If file is blank or does not exist at the default secrets path, a new file will be created.
     - If any environment variables exist in the form of TAP_MY_TAP_my_setting, a new file
    will be created which contains these settings.
     - If the default file exists and environment variables also exist, the temp file will
    contain the default file values along with the environment variable overrides.
    """
    secrets_path = config_dir or _get_secrets_dir()
    default_path = f"{secrets_path}/{plugin_name}-config.json"
    tmp_path = f"{secrets_path}/tmp/{plugin_name}-config.json"
    use_tmp_file = False
    if io.file_exists(default_path):
        json_text = io.get_text_file_contents(default_path)
        conf_dict = json.loads(json_text)
    else:
        conf_dict = {}
        use_tmp_file = True
    for k, v in os.environ.items():
        prefix = f"{plugin_name.replace('-', '_').upper()}_"
        if k.startswith(prefix):
            setting_name = k.split(prefix)[1]
            conf_dict[setting_name] = v
            use_tmp_file = True
    if use_tmp_file:
        io.create_folder(str(Path(tmp_path).parent))
        io.create_text_file(tmp_path, json.dumps(conf_dict))
        return tmp_path
    else:
        return default_path


@logged(
    "installing '{plugin_name}' as '{alias or plugin_name}' "
    "using 'pip3 install {source or plugin_name}'"
)
def install(plugin_name, source=None, alias=None):
    source = source or plugin_name
    alias = alias or plugin_name

    venv_dir = os.path.join(VENV_ROOT, alias)
    install_path = os.path.join(INSTALL_ROOT, alias)
    if io.file_exists(install_path):
        response = input(
            f"The file '{install_path}' already exists. "
            f"Are you sure you want to replace this file? [y/n]"
        )
        if not response.lower() in ["y", "yes"]:
            raise RuntimeError(f"File already exists '{install_path}'.")
        io.delete_file(install_path)
    jobs.run_command(f"python3 -m venv {venv_dir}")
    jobs.run_command(f"{os.path.join(venv_dir ,'bin', 'pip3')} install {source}")
    jobs.run_command(f"ln -s {venv_dir}/bin/{plugin_name} {install_path}")


@logged("running discovery on '{tap_name}'")
def discover(tap_name, config_file=None, catalog_dir=None):
    if env.is_windows() or env.is_mac():
        _rerun_dockerized(tap_name)
        return
    config_file = config_file or _get_config_file(f"tap-{tap_name}")
    catalog_dir = catalog_dir or _get_catalog_output_dir(tap_name)
    catalog_file = f"{catalog_dir}/{tap_name}-catalog-raw.json"
    io.create_folder(catalog_dir)
    jobs.run_command(f"tap-{tap_name} --config {config_file} --discover > {catalog_file}")


@logged("Updating plan file for 'tap-{tap_name}'")
def plan(tap_name, taps_dir=None, config_file=None, config_dir=None, rescan=None):
    """
    Perform all actions necessary to prepare (plan) for a tap execution:
     1. Scan (discover) the source system metadata (if catalog missing or `rescan=True`)
     2. Apply filter rules from `select_file` and create human-readable `plan.yml` file to
        describe planned inclusions/exclusions.
     2. Create a new `catalog-selected.json` file which applies the plan file and which
        can be used by the tap to run data extractions.
    """
    if env.is_windows() or env.is_mac():
        _rerun_dockerized(tap_name)
        return
    taps_dir = _get_taps_dir(taps_dir)
    config_file = config_file or _get_config_file(
        f"tap-{tap_name}", config_dir=config_dir
    )
    catalog_dir = _get_catalog_output_dir(tap_name)
    catalog_file = f"{catalog_dir}/{tap_name}-catalog-raw.json"
    selected_catalog_file = f"{catalog_dir}/{tap_name}-catalog-selected.json"
    plan_file = _get_plan_file(tap_name, taps_dir)
    if (
        io.file_exists(catalog_file)
        and io.get_text_file_contents(catalog_file).strip() == ""
    ):
        logging.info(f"Cleaning up empty catalog file: {catalog_file}")
        io.delete_file(catalog_file)
    if rescan or not io.file_exists(catalog_file):
        discover(tap_name, config_file, catalog_dir)
    select_file = _get_select_file(taps_dir)
    select_rules = [
        line.split("#")[0].rstrip()
        for line in io.get_text_file_contents(select_file).splitlines()
        if line.split("#")[0].rstrip()
    ]
    matches = {}
    excluded_tables = []
    for table_object in _get_catalog_table_objects(catalog_file):
        table_name = table_object["stream"]
        table_match_text = f"{tap_name}.{table_name}"
        if _table_match_check(table_match_text, select_rules):
            matches[table_name] = {}
            for col_object in _get_catalog_table_columns(table_object):
                col_name = col_object
                col_match_text = f"{tap_name}.{table_name}.{col_name}"
                matches[table_name][col_name] = _col_match_check(
                    col_match_text, select_rules
                )
        else:
            excluded_tables.append(table_name)
    sorted_tables = sorted(matches.keys())
    file_text = ""
    file_text += f"selected_tables:\n"
    for table in sorted_tables:
        included_cols = [col for col, selected in matches[table].items() if selected]
        ignored_cols = [col for col, selected in matches[table].items() if not selected]
        file_text += f"{'  ' * 1}{table}:\n"
        file_text += f"{'  ' * 2}selected_columns:\n"
        for col in included_cols:
            file_text += f"{'  ' * 2}- {col}\n"
        if ignored_cols:
            file_text += f"{'  ' * 2}ignored_columns:\n"
            for col in ignored_cols:
                file_text += f"{'  ' * 2}- {col}\n"
    if excluded_tables:
        file_text += f"ignored_tables:\n"
        for table in sorted(excluded_tables):
            file_text += f"{'  ' * 1}- {table}\n"
    io.create_text_file(plan_file, file_text)
    _create_selected_catalog(
        tap_name,
        plan_file=plan_file,
        full_catalog_file=catalog_file,
        output_file=selected_catalog_file,
    )


@logged("syncing '{table_name or 'all tables'}' from '{tap_name}' to '{target_name}'")
def sync(
    tap_name,
    target_name="csv",
    table_name="*",
    taps_dir=None,
    *,
    config_file=None,
    config_dir=None,
    catalog_dir=None,
    target_config_file=None,
    rescan=False,
    state_file=None,
    dockerized: bool = None,
):
    """ Run a tap sync. If table_name is omitted, all sources will be extracted. """
    if dockerized is None:
        if env.is_windows() or env.is_mac():
            logging.info(
                "The 'dockerized' argument is not set when running either Windows or OSX."
                "Attempting to run sync from inside docker."
            )
            _rerun_dockerized(tap_name, target_name)
            return
    taps_dir = _get_taps_dir(taps_dir)
    select_file = _get_select_file(taps_dir)
    config_file = config_file or _get_config_file(f"tap-{tap_name}", config_dir)
    target_config_file = target_config_file or _get_config_file(
        f"target-{target_name}", config_dir
    )
    catalog_dir = catalog_dir or _get_catalog_output_dir(tap_name)
    full_catalog_file = f"{catalog_dir}/{tap_name}-catalog-selected.json"
    if rescan or select_file or not io.file_exists(full_catalog_file):
        plan(
            tap_name,
            taps_dir=taps_dir,
            config_file=config_file,
            config_dir=catalog_dir,
            rescan=rescan,
        )
    if table_name and table_name != "*":
        catalog_file = f"{catalog_dir}/{tap_name}-{table_name}-catalog.json"
        state_file = state_file or f"{catalog_dir}/{table_name}-state.json"
        _create_single_table_catalog(
            tap_name=tap_name,
            table_name=table_name,
            full_catalog_file=full_catalog_file,
            output_file=catalog_file,
        )
    else:
        catalog_file = full_catalog_file
        state_file = state_file or f"{catalog_dir}/full-extract-state.json"
    sync_cmd = (
        f"tap-{tap_name} --config {config_file} --catalog {catalog_file} | "
        f"target-{target_name} --config {target_config_file} "
        f">> {state_file}"
    )
    jobs.run_command(sync_cmd)
    # # tail -1 state.json > state.json.tmp && mv state.json.tmp state.json


def _get_catalog_table_objects(catalog_file):
    catalog_full = json.loads(Path(catalog_file).read_text())
    table_objects = catalog_full["streams"]
    return sorted(table_objects, key=lambda x: x["stream"])


def _get_catalog_table_columns(table_object):
    return table_object["schema"]["properties"].keys()


def _table_match_check(match_text: str, select_rules: list):
    selected = False
    for rule in select_rules:
        result = _check_table_rule(match_text, rule)
        if result == True:
            selected = True
        elif result == False:
            selected = False
    return selected


def _col_match_check(match_text: str, select_rules: list):
    selected = False
    for rule in select_rules:
        result = _check_column_rule(match_text, rule)
        if result == True:
            selected = True
        elif result == False:
            selected = False
    return selected


def _is_match(value, pattern):
    if not pattern:
        return None
    if value.lower() == pattern.lower():
        return True
    if pattern == "*":
        return True
    re_pattern = None
    if "/" in pattern:
        if pattern[0] == "/" and pattern[-1] == "/":
            re_pattern = pattern[1:-1]
            re_pattern = f"\\b{re_pattern}\\b"
            # logging.info(f"Found regex pattern: {pattern}")
    elif "*" in pattern:
        # logging.info(f"Found wildcard pattern: {pattern}")
        re_pattern = pattern.replace("*", ".*")
    if re_pattern:
        # logging.info(f"Checking regex pattern: {re_pattern}")
        result = re.search(re_pattern.lower(), value.lower())
        if result:
            # logging.info(f"Matched regex pattern '{re_pattern}' on '{value}'")
            return True
    return False


# @logged("checking '{match_text}' against rule '{rule_text}'")
def _check_column_rule(match_text: str, rule_text: str):
    """ Checks rule. Returns True to include, False to exclude, or None if not a match """
    if rule_text[0] == "!":
        match_result = False  # Exclude if matched
        rule_text = rule_text[1:]
    else:
        match_result = True  # Include if matched
    rule_text = rule_text.replace("**.", "*.*.")
    tap_match, table_match, column_match = (
        rule_text.split(".")[0],
        rule_text.split(".")[1],
        ".".join(rule_text.split(".")[2:]),
    )
    if not _is_match(match_text.split(".")[0], tap_match):
        return None
    if not _is_match(match_text.split(".")[1], table_match):
        return None
    if not _is_match(match_text.split(".")[2], column_match):
        return None
    # logging.info(
    #     f"Column '{match_text}' matched column filter '{column_match}' in '{rule_text}'"
    # )
    return match_result


def _check_table_rule(match_text: str, rule_text: str):
    """ Checks rule. Returns True to include, False to exclude, or None if not a match """
    if rule_text[0] == "!":
        match_result = False  # Exclude if matched
        rule_text = rule_text[1:]
    else:
        match_result = True  # Include if matched
    rule_text = rule_text.replace("**.", "*.*.")
    if "*.*." in rule_text:
        return None  # Global column rules do not apply to tables
    if match_result == True and rule_text[-2:] == ".*":
        rule_text = rule_text[:-2]
    if len(rule_text.split(".")) > 2:
        return None  # Column rules do not apply to tables
    tap_name, table_name = (match_text.split(".")[0], ".".join(match_text.split(".")[1:]))
    tap_rule, table_rule = (rule_text.split(".")[0], ".".join(rule_text.split(".")[1:]))
    if not _is_match(tap_name, tap_rule):
        return None
    if not _is_match(table_name, table_rule):
        return None
    # logging.info(
    #     f"Table '{match_text}' matched table filter '{table_rule}' in '{rule_text}'"
    # )
    return match_result


@logged(
    "selecting catalog metadata "
    "from '{tap_name}' source catalog file: {full_catalog_file}"
)
def _create_selected_catalog(
    tap_name, plan_file=None, full_catalog_file=None, output_file=None
):
    catalog_dir = _get_catalog_output_dir(tap_name)
    source_catalog_path = full_catalog_file or os.path.join(
        catalog_dir, "catalog-raw.json"
    )
    output_file = output_file or os.path.join(catalog_dir, f"selected-catalog.json")
    catalog_full = json.loads(Path(source_catalog_path).read_text())
    full_table_list = sorted([tbl["stream"] for tbl in catalog_full["streams"]])
    plan_file = plan_file or _get_plan_file(tap_name)
    plan = yaml.safe_load(io.get_text_file_contents(plan_file))
    included_table_objects = []
    for tbl in catalog_full["streams"]:
        stream_name = tbl["stream"]
        if stream_name in plan["selected_tables"].keys():
            _select_table(tbl)
            for col_name in _get_catalog_table_columns(tbl):
                col_selected = (
                    col_name in plan["selected_tables"][stream_name]["selected_columns"]
                )
                _select_table_column(tbl, col_name, col_selected)
            included_table_objects.append(tbl)
    catalog_new = {"streams": included_table_objects}
    with open(output_file, "w") as f:
        json.dump(catalog_new, f, indent=2)


def _select_table(tbl: object):
    for metadata in tbl["metadata"]:
        if len(metadata["breadcrumb"]) == 0:
            metadata["metadata"]["selected"] = True


def _select_table_column(tbl: object, col_name: str, selected: bool):
    for metadata in tbl["metadata"]:
        if (
            len(metadata["breadcrumb"]) >= 2
            and metadata["breadcrumb"][0] == "properties"
            and metadata["breadcrumb"][1] == col_name
        ):
            metadata["metadata"]["selected"] = selected
            return
    tbl["metadata"].append(
        {"breadcrumb": ["properties", col_name], "metadata": {"selected": selected}}
    )
    return


@logged(
    "selecting '{table_name}' catalog metadata "
    "from '{tap_name}' source catalog file: {full_catalog_file}"
)
def _create_single_table_catalog(
    tap_name, table_name, full_catalog_file=None, output_file=None
):
    catalog_dir = _get_catalog_output_dir(tap_name)
    source_catalog_path = full_catalog_file or os.path.join(
        catalog_dir, "catalog-selected.json"
    )
    output_file = output_file or os.path.join(catalog_dir, f"{table_name}-catalog.json")
    included_table_objects = []
    catalog_full = json.loads(Path(source_catalog_path).read_text())
    full_table_list = sorted([tbl["stream"] for tbl in catalog_full["streams"]])
    for tbl in catalog_full["streams"]:
        stream_name = tbl["stream"]
        if stream_name == table_name:
            for metadata in tbl["metadata"]:
                if len(metadata["breadcrumb"]) == 0:
                    metadata["metadata"]["selected"] = True
            included_table_objects.append(tbl)
    catalog_new = {"streams": included_table_objects}
    with open(output_file, "w") as f:
        json.dump(catalog_new, f, indent=2)


def _get_docker_tap_image(tap_alias, target_alias=None):
    if tap_alias.startswith("tap-"):
        tap_alias = tap_alias.replace("tap-", "")
    if not target_alias:
        return f"{BASE_DOCKER_IMAGE}:tap-{tap_alias}"
    else:
        if target_alias.startswith("target-"):
            target_alias = target_alias.replace("target-", "")
        return f"{BASE_DOCKER_IMAGE}:{tap_alias}-to-{target_alias}"


def _rerun_dockerized(tap_alias, target_alias=None):
    cmd = f"s-tap {' '.join(sys.argv[1:])}"
    env = {
        k: v
        for k, v in os.environ.items()
        if (k.startswith("TAP_") or k.startswith("TARGET_"))
    }
    docker_client = docker.from_env()
    image_name = _get_docker_tap_image(tap_alias, target_alias)
    try:
        dockerutils.pull(image_name)
    except Exception as ex:
        logging.warning(f"Could not pull latest Spark image '{image_name}'. {ex}")
    with logged_block(f"running dockerized command '{cmd}' on image '{image_name}'"):

        def _build_docker_run(image, command, environment, working_dir, volumes):
            e_str = " ".join([f"-e {k}={v}" for k, v in environment.items()])
            w_str = f"-w {working_dir}" if working_dir else ""
            v_str = " ".join([f"-v {x}:{y}" for x, y in volumes.items()])
            docker_run_cmd = f"docker run {e_str} {v_str} {w_str} {image} {command}"
            return docker_run_cmd

        volumes = {os.path.abspath("."): "/projects/my-project"}
        DEBUG = True
        if DEBUG:
            container_lib = "/usr/local/lib/python3.8/site-packages/slalom"
            host_lib = "C:\Files\Source\dataops-tools\slalom"
            volumes[host_lib] = container_lib
        docker_run_cmd = _build_docker_run(
            image=image_name,
            command=cmd,
            environment=env,
            working_dir="/projects/my-project",
            volumes=volumes,
        )
        return_code, output_text = jobs.run_command(docker_run_cmd)
    return True


def build_image(tap_or_plugin_alias, target_alias=None, push=False, pre=False):
    name, source, alias = _get_plugin_info(f"tap-{tap_or_plugin_alias}")
    _build_plugin_image(name, source=source, alias=alias, push=push, pre=pre)
    if target_alias:
        name, source, alias = _get_plugin_info(f"target-{target_alias}")
        _build_plugin_image(name, source=source, alias=alias, push=push, pre=pre)
        _build_composite_image(
            tap_alias=tap_or_plugin_alias, target_alias=target_alias, push=push, pre=pre
        )


def _get_plugins_list(plugins_index=None):
    plugins_index = plugins_index or SINGER_PLUGINS_INDEX
    if not io.file_exists(plugins_index):
        raise RuntimeError(
            f"No file found at '{plugins_index}'."
            "Please set SINGER_PLUGINS_INDEX and try again."
        )
    yml_doc = yaml.safe_load(io.get_text_file_contents(plugins_index))
    taps = yml_doc["singer-taps"]
    list_of_tuples = []
    taps = yml_doc["singer-taps"]
    targets = yml_doc["singer-targets"]
    plugins = taps + targets
    for plugin in plugins:
        list_of_tuples.append(
            (plugin["name"], plugin.get("source", None), plugin.get("alias", None))
        )
    return list_of_tuples


def _build_all_standalone(source_image=None, plugins_index=None, push=False, pre=False):
    plugins = _get_plugins_list(plugins_index)
    for name, source, alias in plugins:
        image_name = _build_plugin_image(
            name,
            source=source,
            alias=alias,
            source_image=source_image,
            push=push,
            pre=pre,
        )


def _get_plugin_info(id, plugins_index=None):
    plugins = _get_plugins_list(plugins_index)
    for name, source, alias in plugins:
        if (alias or name) == id:
            return (name, source, alias)
    raise ValueError(f"Could not file a plugin called '{id}'")


def _build_all_composite(source_image=None, plugins_index=None, push=False, pre=False):
    plugins = _get_plugins_list(plugins_index)
    for tap_name, tap_source, tap_alias in plugins:
        tap_alias = tap_alias or tap_name
        for target_name, target_source, target_alias in plugins:
            target_alias = target_alias or target_name
            if tap_alias.startswith("tap-") and target_alias.startswith("target-"):
                image_name = _build_composite_image(
                    tap_alias, target_alias, push=push, pre=pre
                )


def _build_plugin_image(
    plugin_name, source, alias, source_image=None, push=False, pre=False
):
    source = source or plugin_name
    alias = alias or plugin_name
    image_name = f"{IMAGE_BASE}:{alias}"
    build_cmd = f"docker build"
    if source_image:
        build_cmd += f" --build-arg source_image={source_image}"
    if pre:
        build_cmd += f" --build-arg prerelease=true"
        image_name += "--pre"
    build_cmd += (
        f" --build-arg PLUGIN_NAME={plugin_name}"
        f" --build-arg PLUGIN_SOURCE={source}"
        f" --build-arg PLUGIN_ALIAS={alias}"
        f" -t {image_name}"
        f" -f plugin.Dockerfile"
        f" ."
    )
    jobs.run_command(build_cmd)
    if push:
        _push(image_name)
    return image_name


def _build_composite_image(tap_alias, target_alias, push=False, pre=False):
    if tap_alias.startswith("tap-"):
        tap_alias = tap_alias.replace("tap-", "", 1)
    if target_alias.startswith("target-"):
        target_alias = target_alias.replace("target-", "", 1)
    image_name = f"{IMAGE_BASE}:{tap_alias}-to-{target_alias}"
    build_cmd = f"docker build"
    if pre:
        build_cmd += " --build-arg source_image_suffix=--pre"
        image_name += "--pre"
    build_cmd += (
        f" --build-arg tap_alias={tap_alias}"
        f" --build-arg target_alias={target_alias}"
        f" -t {image_name}"
        f" -f tap-to-target.Dockerfile"
        f" ."
    )
    jobs.run_command(build_cmd)
    if push:
        _push(image_name)
    return image_name


def _push(image_name):
    jobs.run_command(f"docker push {image_name}")


def build_all_images(push=False, pre=False):
    """
    Build all images.
    :param push: Push images after building
    :param pre: Create and publish pre-release builds
    """
    _build_all_standalone(push=push, pre=pre)
    _build_all_composite(push=push, pre=pre)


def main():
    fire.Fire(
        {
            "install": install,
            "discover": discover,
            "plan": plan,
            "sync": sync,
            "build_image": build_image,
            "build_all_images": build_all_images,
        }
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import datetime
from distutils.util import strtobool
import hashlib
import inspect
import os
from pathlib import Path
import platform
import subprocess
import sys
import tempfile
import time

import fire

sys.path.append("../src/")
if __name__ == "__main__" and __package__ is None:
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(os.path.join(repo_dir, "src"))

from slalom.dataops.logs import logged, get_logger, fstr, flush_all_output
from slalom.dataops import io


logging = get_logger("slalom.dataops")

DATA_REPO_ROOT = "s3://propensity-to-buy/data"
if "ARTIFACTS_ROOT" not in os.environ:
    os.environ["ARTIFACTS_ROOT"] = tempfile.mkdtemp()
ARTIFACTS_ROOT = os.environ["ARTIFACTS_ROOT"]


def init_batch_id():
    if "BATCH_ID" in os.environ:
        batch_id = os.environ["BATCH_ID"]
        logging.info(f"BATCH_ID:\t{batch_id} (detected)")
    else:
        batch_id = "{:%Y%m%d.%H%M%S}".format(datetime.datetime.now())
        os.environ["BATCH_ID"] = batch_id  # Only shared with child processes
        logging.info(f"BATCH_ID:\t{batch_id} (newly created)")
    return batch_id


# If debugging or if running in CI/CD, only do a dry run (much faster)
BATCH_ID = init_batch_id()
DEV_MODE = strtobool(os.environ.get("DEV_MODE", "0"))
DRY_RUN_MODE = DEV_MODE or ("CI" in os.environ)


def get_project_steps(path, as_dag=False):
    """ Returns the project steps either as a dag or as an ordered list. """
    if as_dag:
        raise NotImplementedError("DAG feature not yet implemented")
    else:
        return


def get_all_script_files(*scripts_folders):
    for folder in scripts_folders:
        for dirpath, _, filenames in os.walk(folder):
            for f in sorted(filenames, key=lambda x: x.replace("_", " ")):
                if ".disabled" not in f and (
                    f.lower().endswith(".r")
                    or f.lower().endswith(".py")
                    or f.lower().endswith(".sql")
                    or f.lower().endswith(".ipynb")
                ):
                    yield os.path.join(dirpath, f)


@logged(
    "replicating cache from '{source_folder}' to '{target_folder}'",
    success_detail="{len(result)} files copied",
)
def replicate_cache(source_folder, target_folder):
    if source_folder != target_folder:
        copied_files = []
        source_files = io.list_s3_files(os.path.join(source_folder, ""))
        for source_file in source_files:
            if not source_file in [
                source_folder + "/_SUCCESS",
                source_folder + "/",
                source_folder,
            ]:
                if source_folder not in source_file:
                    logging.warning(
                        f"Problem detected. Folder path '{source_folder}' not "
                        f"contained in '{source_file}'"
                    )
                target_file = source_file.replace(source_folder, target_folder)
                if target_folder not in target_file:
                    logging.warning(
                        f"Problem detected. Folder path '{target_folder}' not "
                        f"contained in '{target_file}'"
                    )
                logging.debug(f"Replicating file cache: {source_file}->{target_file}")
                io.copy_s3_file(source_file, target_file)
                copied_files.append(target_file)
        return copied_files
    else:
        logging.debug(
            "Source and target are the same. Skipping '{source_folder}' replication."
        )


@logged("'{script_file_path}' script job", buffer_lines=2)
def generate_script_output(
    script_file_path,
    parent_hash,
    batch_output_dir,
    use_cache=True,
    save_cache=True,
    replicate_cache_if_skipped=True,
    schema_only=False,
):
    """ Run the job or if code hashes match, clone resources and skip rebuild """
    file_type = script_file_path.split(".")[-1].lower()
    cmd = None
    cli_shell = False
    if file_type == "md":
        pass
    elif ".disabled" in script_file_path:
        logging.info(f"Skipping disabled file: {script_file_path}")
    elif file_type == "py":
        logging.debug(f"Identified Python script: '{script_file_path}'")
        cmd = [sys.executable, script_file_path]
    elif file_type == "r":
        logging.debug(f"Identified R script: '{script_file_path}'")
        cmd = ["Rscript", script_file_path]
        cli_shell = False
    elif file_type == "sql":
        logging.debug(f"Identified SQL script: '{script_file_path}'")
        cmd = [
            sys.executable,
            "project/scripts/ptb_model_scripts/01_feat_eng_spark.py",
            "--sql_file_path",
            script_file_path,
        ]
    else:
        raise NotImplementedError(
            f"Script type not supported: '*.{file_type}' in script '{script_file_path}'"
        )
    if cmd:
        start_time = time.time()
        log_file_name = f"{os.path.basename(script_file_path)}.log"
        log_file_path = os.path.join(ARTIFACTS_ROOT, log_file_name)
        new_running_hash = get_appended_code_hash(parent_hash, script_file_path)
        parent_cache_folder, new_cache_folder = None, None
        if use_cache:
            parent_cache_folder = get_cache_folder_path(parent_hash)
        if save_cache:
            new_cache_folder = get_cache_folder_path(new_running_hash)
        if use_cache and io.file_exists(os.path.join(new_cache_folder, "_SUCCESS")):
            logging.info(
                f"Skipping '{script_file_path}' execution "
                f"and using cache from {new_cache_folder}"
            )
            prev_log_path = os.path.join(new_cache_folder, "logs", log_file_name)
            if io.file_exists(prev_log_path):
                io.download_s3_file(prev_log_path, log_file_path)
                flush_all_output()
                with open(log_file_path, "rU", encoding="utf-8") as prev_log:
                    sys.stdout.write(
                        "\n".join(["|| " + l.rstrip() for l in prev_log.readlines()])
                    )
                flush_all_output()
            if replicate_cache_if_skipped:
                replicate_cache(new_cache_folder, batch_output_dir)
            else:
                logging.info(
                    "Skipping execution and replication of already-cached script output: "
                    f"'{script_file_path}' (cache folder: '{new_cache_folder}')"
                )
        else:
            if (
                use_cache
                and parent_cache_folder
                and io.file_exists(os.path.join(parent_cache_folder, "_SUCCESS"))
            ):
                logging.debug(
                    f"Found usable cache for '{script_file_path}' "
                    f"(hash={new_running_hash})...\n\n"
                )
                replicate_cache(parent_cache_folder, new_cache_folder)
                os.environ["OUTPUT_DIR_OVERRIDE"] = new_cache_folder
                work_dir = new_cache_folder
            else:  # not using cache
                use_cache = False
                work_dir = batch_output_dir
                if "OUTPUT_DIR_OVERRIDE" in os.environ:
                    del os.environ["OUTPUT_DIR_OVERRIDE"]  # ?Is this needed?
            logging.info(
                f"Running script '{script_file_path}' (hash={new_running_hash})...\n\n"
                f"{'-' * 80}\n"
                f"{'-' * 80}\n\n"
            )
            return_code, output_text = run_command(
                cmd,
                raise_error=True,
                log_file_path=log_file_path,
                echo=True,
                shell=cli_shell,
            )
            io.create_s3_text_file(
                os.path.join(work_dir, "logs", log_file_name), contents=output_text
            )
            logging.debug(f"Script execution completed.")
            if use_cache:
                replicate_cache(work_dir, batch_output_dir)
            elif save_cache:
                replicate_cache(work_dir, new_cache_folder)
            if save_cache:
                io.create_s3_text_file(
                    os.path.join(new_cache_folder, "_SUCCESS"), contents=""
                )
        return new_running_hash
    return parent_hash


def _grep(full_text, match_with, insensitive=True, fn=any):
    lines = full_text.splitlines()
    if isinstance(match_with, str):
        match_with = [match_with]
    if insensitive:
        return "\n".join(
            [l for l in lines if fn([m.lower() in l.lower() for m in match_with])]
        )
    else:
        return "\n".join([l for l in lines if fn([m in l for m in match_with])])


@logged("running command: {'(hidden)' if hide else cmd}")
def run_command(
    cmd: str,
    working_dir=None,
    echo=True,
    raise_error=True,
    log_file_path: str = None,
    shell=True,
    daemon=False,
    hide=False,
    cwd=None,
    wait_test=None,
    wait_max=None,
):
    """ Run a CLI command and return a tuple: (return_code, output_text) """
    loglines = []
    if working_dir:
        prev_working_dir = os.getcwd()
        os.chdir(working_dir)
    if isinstance(cmd, list):
        pass  # cmd = " ".join(cmd)
    elif platform.system() == "Windows":
        cmd = " ".join(cmd.split())
    else:
        cmd = cmd.replace("\n", " \\\n")
    proc = subprocess.Popen(
        cmd,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
        shell=shell,
        cwd=cwd,
    )
    start_time = time.time()
    if working_dir:
        os.chdir(prev_working_dir)
    if log_file_path:
        logfile = open(log_file_path, "w", encoding="utf-8")
    else:
        logfile = None
    line = proc.stdout.readline()
    flush_all_output()
    while (proc.poll() is None) or line:
        if daemon:
            if wait_max is None and wait_test is None:
                logging.info("Daemon process is launched. Returning...")
                break
            if callable(wait_test) and wait_test(line):
                logging.info(f"Returning. Wait test passed: {line}")
                break
            if wait_max and time.time() >= start_time + wait_max:
                logging.info(
                    f"{line}\nMax timeout expired (wait_max={wait_max})."
                    f" Returning..."
                )
                if callable(wait_test):
                    return_code = 1
                else:
                    return_code = 0
                break
        if line:
            line = line.rstrip()
            loglines.append(line)
            if echo:
                for l in line.splitlines():
                    sys.stdout.write(l.rstrip() + "\r\n")
            if logfile:
                logfile.write(line + "\n")
        else:
            time.sleep(0.5)  # Sleep half a second if no new output
        line = proc.stdout.readline()
    flush_all_output()
    output_text = chr(10).join(loglines)
    if logfile:
        logfile.close()
    if not proc:
        return_code = None
        raise RuntimeError(f"Command failed: {cmd}\n\n")
    else:
        return_code = proc.returncode
        if (
            return_code != 0
            and raise_error
            and ((daemon == False) or (return_code is not None))
        ):
            err_msg = f"Command failed (exit code {return_code}): {cmd}"
            if not echo:
                print_str = output_text
            elif len(output_text.splitlines()) > 10:
                print_str = _grep(
                    output_text, ["error", "exception", "warning", "fail", "deprecat"]
                )
            else:
                print_str = ""
            if print_str:
                err_msg += (
                    f"{'-' * 80}\n"
                    f"SCRIPT ERRORS:\n{'-' * 80}\n"
                    f"{print_str}\n{'-' * 80}\n"
                    f"END OF SCRIPT OUTPUT\n{'-' * 80}"
                )
            raise RuntimeError(err_msg)
    return return_code, output_text


def get_cache_folder_path(code_hash):
    DATA_REPO_ROOT = "s3://propensity-to-buy/data"
    return os.path.join(DATA_REPO_ROOT, "temp/cache", code_hash)


def get_batch_folder_path(batch_id):
    if DRY_RUN_MODE:
        return f"{DATA_REPO_ROOT}/out/dry-run=True/batch={batch_id}"
    else:
        return f"{DATA_REPO_ROOT}/out/batch={batch_id}"


def get_appended_code_hash(prev_code_hash: str, script_file_path: str):
    new_file_hash = hashlib.md5(Path(script_file_path).read_bytes()).hexdigest()
    return hashlib.md5((prev_code_hash + new_file_hash).encode("utf-8")).hexdigest()


@logged("running {len(job_steps)} jobs")
def run_jobs(job_steps, use_cache=True, save_cache=True):
    """ Execute all steps in the provided list or dag """
    if not isinstance(job_steps, list):
        raise NotImplementedError(
            "List expected. (DAG feature not yet implemented.)"
            f"Argument was: {job_steps}"
        )
    # current_file_text = inspect.getsource(inspect.getmodule(inspect.currentframe()))
    # app_version_seed = current_file_text
    app_version_seed = (
        f"Version=1.0.4;"
        f"yyyymmdd={os.environ.get('YYYYMMDD', None)};DryRun={DRY_RUN_MODE}"
    )
    running_code_hash = hashlib.md5(app_version_seed.encode("utf-8")).hexdigest()
    for i, job_step in enumerate(job_steps, 1):
        is_last_job = i == len(job_steps)
        output_dir = get_batch_folder_path(BATCH_ID)
        running_code_hash = generate_script_output(
            script_file_path=job_step,
            parent_hash=running_code_hash,
            batch_output_dir=output_dir,
            use_cache=use_cache,
            save_cache=save_cache,
            replicate_cache_if_skipped=is_last_job,
        )



from contextlib import contextmanager
import datetime
import inspect
import logging
import threading
import queue
import os
import sys
import time

try:
    import psutil
except Exception as ex:
    psutil = None
    logging.warning(
        f"Could not load psutils. Some logging function may not be available. {ex}"
    )


def get_logger(app_name, debug=os.environ.get("DEBUG", None)):
    global DEFAULT_LOGGER

    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(format=log_format)
    logger = logging.getLogger(app_name)
    logger.setLevel(log_level)
    DEFAULT_LOGGER = logger
    return logger


DEFAULT_HEARTBEAT_MINUTES = 5
DEFAULT_LOGGER = get_logger("slalom.dataops.logs")


def duration_to_string(seconds):
    """ Return duration as a concise string. e.g. "32min 3s", "4hr 34min", etc """
    units = ["hr", "min", "s"]
    duration_parts = [
        int(part) for part in str(datetime.timedelta(seconds=int(seconds))).split(":")
    ]
    if duration_parts[0]:
        # if >=1hr, append seconds as tenth of minutes
        # duration_parts[1] = duration_parts[1] + round(duration_parts[2] / 60, 1)
        duration_parts = duration_parts[:2]
    result = " ".join(
        [str(part) + units[x] for x, part in enumerate(duration_parts, 0) if part]
    )
    return result or "0s"


def elapsed_since(start, template="({duration} elapsed)"):
    """ Return a formatted string, e.g. '(HH:MM:SS elapsed)' """
    seconds = time.time() - start
    duration = duration_to_string(seconds=int(seconds))
    return fstr(template, locals())


def flush_all_output():
    """ flush the logging buffers, stderr, and stdout """
    sys.stdout.flush()
    sys.stderr.flush()
    for loghandler in DEFAULT_LOGGER.handlers:
        loghandler.flush()
    sys.stdout.flush()
    sys.stderr.flush()


def convert_mem_units(
    from_val, from_units: str = None, to_units: str = None, sig_digits=None
):
    from_units = from_units or "B"
    _mem_units_map = {
        "B": 1,
        "K": (1024 ** 1),
        "MB": (1024 ** 2),
        "GB": (1024 ** 3),
        "TB": (1024 ** 4),
    }
    num_bytes = from_val * _mem_units_map[from_units]
    return_tuple = not to_units
    if not to_units:
        cutover_factor = 800
        if not to_units in _mem_units_map:
            if num_bytes < 100:  # < 800 K as K
                to_units = "B"
            if num_bytes < cutover_factor * _mem_units_map["K"]:  # < 800 K as K
                to_units = "K"
            elif num_bytes < cutover_factor * _mem_units_map["MB"]:  # < 800 MB as MB
                to_units = "MB"
            elif num_bytes < cutover_factor * _mem_units_map["GB"]:  # < 800 GB as GB
                to_units = "GB"
            else:  # >= 800 TB as TB
                to_units = "TB"
    result = num_bytes * 1.0 / _mem_units_map[to_units]
    if not sig_digits:
        sig_digits = 1 if result >= 10 else 2
    if return_tuple:
        return round(result, sig_digits), to_units
    else:
        return round(result, sig_digits)


def bytes_to_string(num_bytes, units=None):
    """
    Return a string that efficiently represents the number of bytes.
    e.g. "476.4MB", "0.92TB", etc.
    """
    new_value, units = convert_mem_units(num_bytes, from_units="B", to_units=None)
    return f"{new_value}{units}"


def ram_usage_string(process_id=None):
    """
    Return a string representing the amount and percentage of memory used by this process.
    """
    if not psutil:
        return "(unknown mem usage - missing psutil library)"
    process = psutil.Process(process_id or os.getpid())
    amount = bytes_to_string(process.memory_info().rss)
    percent = process.memory_percent()
    return f"(mem usage: {amount}, {percent:,.1f}%)"


def cpu_usage_string(process_id=None):
    """
    Return a string representing the amount and percentage of memory used by this process.
    """
    if not psutil:
        return "(unknown CPU usage - missing psutil library)"
    process = psutil.Process(process_id or os.getpid())
    return f"(CPU {process.cpu_percent(interval=0.2)}%)"


def _get_printable_context(context: dict = None, as_str=True):
    """ Return a string or dict, obfuscating names that look like keys """
    printable_dict = {
        k: (
            v
            if not any(
                [
                    "secret" in k.lower(),
                    "pwd" in k.lower(),
                    "pass" in k.lower(),
                    "access.key" in k.lower(),
                ]
            )
            else "****"
        )
        for k, v in context.items()
        if k != "__builtins__"
    }
    if as_str:
        return "\n".join([f"\t{k}:\t{v}" for k, v in printable_dict.items()])
    else:
        return printable_dict


def _caller_and_lineno():
    caller = inspect.getframeinfo(inspect.stack()[1][0])
    return f"{caller.filename}:{caller.lineno}"


def fstr(fstring_text, locals, globals=None):
    """
    Dynamically evaluate the provided fstring_text

    Sample usage:
        format_str = "{i}*{i}={i*i}"
        i = 2
        fstr(format_str, locals()) # "2*2=4"
        i = 4
        fstr(format_str, locals()) # "4*4=16"
        fstr(format_str, {"i": 12}) # "10*10=100"
    """
    locals = locals or {}
    globals = globals or {}
    result = eval(f'f"{fstring_text}"', locals, globals)
    return result


def heartbeat_printer(
    desc_text,
    msg_queue: queue.Queue,
    interval,
    show_memory=True,
    show_cpu=None,
    start_time=None,
):
    start_time = start_time or time.time()
    show_memory = show_memory if show_memory is not None else True
    show_cpu = show_cpu if show_cpu is not None else show_memory
    time.sleep(interval)
    while msg_queue.empty():
        elapsed_str = elapsed_since(start_time, template="({duration} and counting...)")
        msg = f"Still {desc_text} {elapsed_str}"
        if show_cpu:
            msg += cpu_usage_string(process_id=None)
        if show_memory:
            msg += ram_usage_string(process_id=None)
        DEFAULT_LOGGER.info(msg)
        time.sleep(interval)


@contextmanager
def logged_block(
    desc_text,
    start_msg="Beginning {desc_text}...",
    success_msg="Completed {desc_text} {success_detail}  {elapsed}",
    success_detail="",
    show_memory=None,
    heartbeat_interval=DEFAULT_HEARTBEAT_MINUTES * 60,
    **kwargs,
):
    """
    Times and logs the execution inside a with block.

    Sample usage:

        with logged_block("running '{job.name}' job", job=job_obj):
            do_job(job)
    """
    start = time.time()
    context_dict = locals().copy()
    context_dict.update(kwargs)
    if start_msg:
        if show_memory:
            start_msg = start_msg + (" " * 15) + ram_usage_string()
        DEFAULT_LOGGER.info(fstr(start_msg, locals=context_dict))
    msg_queue = None
    if heartbeat_interval:
        msg_queue = queue.Queue()
        heartbeat = threading.Thread(
            target=heartbeat_printer,
            args=[],
            kwargs={
                "desc_text": desc_text,
                "msg_queue": msg_queue,
                "interval": heartbeat_interval,
                "show_memory": show_memory,
            },
        )
        heartbeat.daemon = True
        heartbeat.start()
    yield
    if heartbeat:
        try:
            msg_queue.put("cancel")
        except Exception as ex:
            DEFAULT_LOGGER.exception("Failed to kill heartbeat log. {ex}")
    context_dict["elapsed"] = elapsed_since(start)
    if success_msg:
        if show_memory:
            success_msg = success_msg + ram_usage_string()
        DEFAULT_LOGGER.info(fstr(success_msg, locals=context_dict))


class logged(object):
    """
    Decorator class for logging function start, completion, and elapsed time.

    Sample usage:
        @logged()
        def my_func_a():
            pass

        @logged(log_fn=logging.debug)
        def my_func_b():
            pass

        @logged("doing a thing")
        def my_func_c():
            pass

        @logged("doing a thing with {foo_obj.name}")
        def my_func_d(foo_obj):
            pass

        @logged("doing a thing with '{custom_kwarg}'", custom_kwarg="foo")
        def my_func_d(foo_obj):
            pass
    """

    def __init__(
        self,
        desc_text="{fn.__name__}() for '{desc_detail}'",
        desc_detail="",
        start_msg="Beginning {desc_text}...",
        success_msg="Completed {desc_text} {elapsed} ({success_detail})",
        success_detail="",
        buffer_lines=0,
        log_fn=None,
        **addl_kwargs,
    ):
        """ All arguments optional """
        log_fn = log_fn or DEFAULT_LOGGER.info
        self.default_context = addl_kwargs.copy()  # start with addl. args
        self.default_context.update(locals())  # merge all constructor args
        self.buffer_lines = buffer_lines

    def print_buffer(self):
        if self.buffer_lines:
            nl = "\n"
            flush_all_output()
            sys.stdout.write(f"\n\n{('-' * 80 + nl) * self.buffer_lines}\n\n")
            flush_all_output()

    def __call__(self, fn):
        """ Call the decorated function """

        def wrapped_fn(*args, **kwargs):
            """
            The decorated function definition. Note that the log needs access to
            all passed arguments to the decorator, as well as all of the function's
            native args in a dictionary, even if args are not provided by keyword.
            If start_msg is None or success_msg is None, those log entries are skipped.
            """

            def re_eval(context_dict, context_key: str):
                """
                Evaluate the f-string in context_dict[context_key],
                store back the result
                """
                try:
                    context_dict[context_key] = fstr(
                        context_dict[context_key], locals=context_dict
                    )
                except Exception as ex:
                    DEFAULT_LOGGER.warning(
                        f"Error evaluating '{context_key}' "
                        f"({context_dict.get(context_key, '(missing)')})"
                        f": '{ex}' with context: '{_get_printable_context(context_dict)}'"
                    )

            start = time.time()
            fn_context = self.default_context.copy()
            fn_context["fn"] = fn
            fn_context["elapsed"] = None
            argspec = inspect.getfullargspec(fn)
            # DEFAULT_LOGGER.info(f"argspec: {argspec}")
            if argspec.defaults:
                # DEFAULT_LOGGER.info(
                #     f"attempting to set defaults: {list(enumerate(argspec.defaults, 1))}"
                # )
                for i, v in enumerate(reversed(argspec.defaults), 1):
                    fn_context[argspec.args[-1 * i]] = v
            if argspec.kwonlydefaults:
                fn_context.update(dict(argspec.kwonlydefaults))
            fn_arg_names = argspec.args.copy()
            # DEFAULT_LOGGER.info(f"args: {fn_arg_names}")
            if argspec.varargs is not None:
                # unnamed orderd args
                fn_context[argspec.varargs] = args
            else:
                for x, arg_value in enumerate(args, 0):
                    # DEFAULT_LOGGER.info(
                    #     f"Attempting to set: fn_arg_names[{x}] = {arg_value}"
                    # )
                    fn_context[fn_arg_names[x]] = arg_value
            fn_context.update(kwargs)
            desc_detail_fn = None
            log_fn = fn_context["log_fn"]
            # If desc_detail is callable, evaluate dynamically (both before and after)
            if callable(fn_context["desc_detail"]):
                desc_detail_fn = fn_context["desc_detail"]
                fn_context["desc_detail"] = desc_detail_fn()
            # Re-evaluate any decorator args which are fstrings
            re_eval(fn_context, "desc_detail")
            re_eval(fn_context, "desc_text")
            # Remove 'desc_detail' if blank or unused
            fn_context["desc_text"] = fn_context["desc_text"].replace("'' ", "")
            re_eval(fn_context, "start_msg")
            if fn_context["start_msg"]:
                self.print_buffer()
                log_fn(fn_context["start_msg"])  # log start of execution
            result = fn(*args, **kwargs)
            if fn_context["success_msg"]:  # log the end of execution
                if callable(fn_context["success_msg"]):
                    fn_context["success_msg"] = fn_context["success_msg"]()
                fn_context["result"] = result
                if desc_detail_fn:  # If desc_detail callable, then reevaluate
                    fn_context["desc_detail"] = desc_detail_fn()
                fn_context["elapsed"] = elapsed_since(start)
                re_eval(fn_context, "success_detail")
                re_eval(fn_context, "success_msg")
                log_fn(fn_context["success_msg"].replace(" ()", ""))
                self.print_buffer()
            return result

        wrapped_fn.__doc__ = fn.__doc__  # Use docstring from inner function.
        return wrapped_fn



import os as _os
import tempfile as _tempfile
from pathlib import Path as _Path
import shutil as _shutil

import fire as _fire

from slalom.dataops import logs as _logs
from slalom.dataops import jobs as _jobs

_LOGGER = _logs.get_logger("slalom.dataops")


def _warn_failed_import(library_name, install_hint):
    _LOGGER.warning(
        f"Could not load '{library_name}' library. Some functionality may be disabled. "
        f"Please confirm '{library_name}' is installed or install via '{install_hint}'."
    )


# try:
#     import pandas as pd
# except Exception as ex:
#     pd = None
#     _warn_failed_import("pandas", "pip install pandas")
try:
    import boto3 as _boto3
except Exception as ex:
    _boto3 = None
    _warn_failed_import("boto3", "pip install boto3")
try:
    import s3fs as _s3fs
except Exception as ex:
    _s3fs = None
    _warn_failed_import("s3fs", "pip install s3fs")

_adls = None
_adls_creds = None
try:
    from azure.datalake import store as _adls
except Exception as ex:
    _LOGGER.warning(f"Azure libraries not loaded: {ex}")
if _adls:
    try:
        _adls_creds = _adls.lib.auth(
            tenant_id=_os.environ["AZURE_TENANT_ID"],
            client_secret=_os.environ["AZURE_CLIENT_SECRET"],
            client_id=_os.environ["AZURE_CLIENT_ID"],
            resource="https://datalake.azure.net/",
        )
    except Exception as ex:
        _LOGGER.warning(f"Azure creds not loaded: {ex}")


_SAFE_PATHS = []

# Enforcement of write safety
def set_safe_paths(list_of_paths):
    """ Set list of safe writeable paths """
    global _SAFE_PATHS
    _SAFE_PATHS = list_of_paths


_USE_SCRATCH_DIR = _os.environ.get("USE_SCRATCH_DIR", False)
_scratch_dir = None
_tmpdir = None


def get_temp_dir():
    global _tmpdir

    if not _tmpdir:
        _tmpdir = _tempfile.mkdtemp()
    return _tmpdir


def get_scratch_dir():
    global _scratch_dir

    if not _scratch_dir:
        _scratch_dir = get_temp_dir()
        _os.environ["SCRATCH_DIR"] = _scratch_dir
    return _scratch_dir


# Helper functions
def _check_one_or_all(string_or_list, fn, agg_fn=all):
    if isinstance(string_or_list, str):
        string = string_or_list
        return fn(string)
    else:
        return agg_fn([fn(string) for string in string_or_list])


# Path parsing
def cleanup_filepath(filepath):
    if "%USERPROFILE%" in filepath and "USERPROFILE" in _os.environ:
        filepath = filepath.replace("%USERPROFILE%", _os.environ["USERPROFILE"])
    filepath = filepath.replace("s3a://", "s3://")
    return filepath


def parse_s3_path(s3_path):
    """ Returns tuple (bucket_name, object_key) """
    path_parts = s3_path.replace("s3://", "").split("/")
    bucket_name, object_key = path_parts[0], "/".join(path_parts[1:])
    return bucket_name, object_key


def parse_adl_path(adl_path):
    """ Returns tuple (store_name, object_key) """
    path_parts = adl_path.replace("adl://", "").split("/")
    store_name = path_parts[0].replace(".azuredatalakestore.net", "")
    object_key = "/".join(path_parts[1:])
    return store_name, object_key


def parse_git_path(git_path):
    """ Returns tuple (repo_url, git_ref, code_path) """
    git_ref = "master"
    repo_url, code_path = git_path.replace("git://", "").split("//")
    if "#" in repo_url:
        repo_url, git_ref = repo_url.split("#")
    repo_url = repo_url.rstrip(".git")
    if not code_path:
        code_path = "/"
    return repo_url, git_ref, code_path


def is_s3(path_or_paths, agg_fn=all):
    check_fn = lambda s: s.startswith("s3")
    return _check_one_or_all(path_or_paths, check_fn, agg_fn)


def is_adl(path_or_paths, agg_fn=all):
    check_fn = lambda s: s.startswith("adl")
    return _check_one_or_all(path_or_paths, check_fn, agg_fn)


def is_git(path_or_paths, agg_fn=all):
    check_fn = lambda s: s.startswith("git")
    return _check_one_or_all(path_or_paths, check_fn, agg_fn)


def is_local(filepath):
    return not any([is_s3(filepath), is_adl(filepath), is_git(filepath)])


def make_local(folder_path):
    if is_local(folder_path):
        return folder_path
    else:
        return download_folder(
            folder_path, local_folder=get_scratch_dir(), as_subfolder=True
        )


def _pick_cloud_function(filepath, s3_fn, adl_fn, git_fn=None, else_fn=None):
    if is_s3(filepath):
        fn = s3_fn
    elif is_adl(filepath):
        fn = adl_fn
    elif is_git(filepath):
        fn = git_fn
    else:
        fn = else_fn
    if not fn:
        raise NotImplemented()
    return fn


# General file operations
def file_exists(filepath):
    filepath = cleanup_filepath(filepath)
    fn = _pick_cloud_function(
        filepath, s3_fn=s3_file_exists, adl_fn=adl_file_exists, else_fn=_os.path.exists
    )
    return fn(filepath)


def s3_file_exists(filepath):
    s3 = _s3fs.S3FileSystem(anon=False)
    return s3.exists(filepath)


def adl_file_exists(filepath):
    store_name, path = parse_adl_path(filepath)
    adl = _adls.core.AzureDLFileSystem(_adls_creds, store_name=store_name)
    return adl.exists(path)


def list_files(file_prefix):
    fn = _pick_cloud_function(
        file_prefix,
        s3_fn=list_s3_files,
        adl_fn=list_adl_files,
        else_fn=lambda prefix: [_os.path.join(prefix, x) for x in _os.listdir(prefix)],
    )
    return fn(file_prefix)


# Function Aliases:
ls = list_files
dir = list_files


@_logs.logged("listing S3 files from '{s3_prefix}'")
def list_s3_files(s3_prefix):
    boto = _boto3.resource("s3")
    bucket_name, folder_key = parse_s3_path(s3_prefix)
    s3_bucket = boto.Bucket(bucket_name)
    file_list = []
    for object_summary in s3_bucket.objects.filter(Prefix=folder_key):
        if object_summary.key[-1] == "/":
            pass  # key is a directory
        else:
            file_list.append(f"s3://{bucket_name}/{object_summary.key}")
    return file_list


def list_adl_files(adl_path_prefix):
    store_name, path_prefix = parse_adl_path(adl_path_prefix)
    adl = _adls.core.AzureDLFileSystem(_adls_creds, store_name=store_name)
    root = adl_path_prefix.split(path_prefix)[0]
    files = adl.walk(path_prefix)
    return [f"{root}{result}" for result in files]


# File copy
def copy_s3_file(s3_source_file, s3_target_file):
    s3 = _s3fs.S3FileSystem(anon=False)
    s3.cp(s3_source_file, s3_target_file)


def copy_adl_file(adl_source_file, adl_target_file):
    raise NotImplemented()


def copy_file(source_file, target_file):
    if is_s3([source_file, target_file]):
        copy_s3_file(source_file, target_file)
    elif is_adl([source_file, target_file]):
        copy_adl_file(source_file, target_file)
    elif is_adl([source_file, target_file], any) and is_adl(
        [source_file, target_file], any
    ):
        tmp_path = download_file(source_file)
        upload_file(tmp_path, target_file)
    else:
        raise _shutil.copyfile(source_file, target_file)


# File deletion
def delete_file(filepath, ignore_missing=True):
    fn = _pick_cloud_function(
        filepath, s3_fn=delete_s3_file, adl_fn=delete_adl_file, else_fn=delete_local_file
    )
    return fn(filepath, ignore_missing=ignore_missing)


def delete_s3_file(s3_filepath, ignore_missing=True):
    if ignore_missing and not s3_file_exists(s3_filepath):
        return None
    boto = _boto3.resource("s3")
    bucket_name, object_key = parse_s3_path(s3_filepath)
    boto.Object(bucket_name, object_key).delete()


def delete_adl_file(adl_filepath, ignore_missing=True):
    if ignore_missing and not adl_file_exists(adl_filepath):
        return None
    store_name, filepath = parse_adl_path(adl_filepath)
    adl = _adls.core.AzureDLFileSystem(_adls_creds, store_name=store_name)
    return adl.rm(filepath)


def delete_local_file(filepath, ignore_missing=True):
    if ignore_missing and not _os.path.exists(filepath):
        return None
    _os.remove(filepath)


# File writes and uploads
def upload_file(local_path, remote_path):
    fn = _pick_cloud_function(
        remote_path, s3_fn=upload_s3_file, adl_fn=upload_adl_file, else_fn=None
    )
    return fn(local_path, remote_path)


def upload_adl_file(local_path, adl_filepath):
    store_name, filepath = parse_adl_path(adl_filepath)
    adl = _adls.core.AzureDLFileSystem(_adls_creds, store_name=store_name)
    _adls.multithread.ADLUploader(
        adl,
        lpath=local_path,
        rpath=filepath,
        nthreads=12,  # WAS: 64, reduced to resolve mem issue: https://github.com/Azure/azure-data-lake-store-python/issues/56
        overwrite=True,
        buffersize=4194304,
        blocksize=4194304,
    )


def upload_s3_file(local_path, s3_filepath):
    s3 = _boto3.client("s3")
    bucket_name, object_key = parse_s3_path(s3_filepath)
    s3.upload_file(local_path, bucket_name, object_key)  # SAFE


def create_s3_text_file(s3_filepath, contents):
    s3 = _boto3.client("s3")
    path_parts = s3_filepath.replace("s3://", "").split("/")
    bucket_name, file_key = path_parts[0], "/".join(path_parts[1:])
    _ = s3.put_object(Bucket=bucket_name, Body=contents, Key=file_key)


def create_text_file(filepath, contents):
    fn = _pick_cloud_function(
        filepath,
        s3_fn=create_s3_text_file,
        adl_fn=None,
        else_fn=lambda filepath, contents: _Path(filepath).write_text(contents),
    )
    return fn(filepath, contents)


# Folder operations
def create_s3_folder(s3_folderpath):
    s3 = _boto3.client("s3")
    bucket_name, folder_key = parse_s3_path(s3_folderpath)
    if not folder_key.endswith("/"):
        folder_key = folder_key + "/"
    _ = s3.put_object(Bucket=bucket_name, Body="", Key=folder_key)


def create_folder(folderpath):
    fn = _pick_cloud_function(
        folderpath,
        s3_fn=create_s3_folder,
        adl_fn=None,
        else_fn=lambda folderpath: _Path(folderpath).mkdir(parents=True, exist_ok=True),
    )
    return fn(folderpath)


# File downloads
def download_file(remote_path, local_path):
    fn = _pick_cloud_function(
        remote_path, s3_fn=download_s3_file, adl_fn=None, else_fn=_shutil.copyfile
    )
    return fn(remote_path, local_path)


@_logs.logged(desc_detail="{s3_path}->{local_path}")
def download_s3_file(s3_path, local_path):
    """ Downloads an S3 file """
    create_folder(_os.path.dirname(local_path))
    boto = _boto3.resource("s3")
    bucket_name, object_key = parse_s3_path(s3_path)
    s3_bucket = boto.Bucket(bucket_name)
    s3_bucket.download_file(object_key, local_path)


def get_text_file_contents(filepath, encoding="utf-8"):
    if not filepath:
        raise ValueError(f"Invalid filepath argument: {filepath}")
    filepath = cleanup_filepath(filepath)
    with open(filepath, "r", encoding=encoding) as f:
        return f.read()


# Folder downloads:
@_logs.logged(desc_detail="{remote_folder}->{local_folder}")
def download_folder(remote_folder, local_folder, as_subfolder=False):
    """ Expects that destination folder does not exist or is empty """
    if as_subfolder:
        local_folder = f"{local_folder}/{_os.path.basename(remote_folder)}"
    fn = _pick_cloud_function(
        remote_folder,
        s3_fn=download_s3_folder,
        adl_fn=None,
        git_fn=download_git_folder,
        else_fn=_download_folder,
    )
    return fn(remote_folder, local_folder)


@_logs.logged(desc_detail="{remote_folder}->{local_folder}")
def _download_folder(remote_folder, local_folder):
    for remote_filepath in list_files(remote_folder):
        sub_path = remote_filepath.split(remote_folder)[1]
        sub_path = sub_path.lstrip("/\\")
        local_filepath = _os.path.join(local_folder, sub_path)
        _LOGGER.info(f"Copying {remote_filepath} to {local_filepath}...")
        download_file(remote_filepath, local_filepath)
    return local_folder


# Alias 'copy_folder'
copy_folder = download_folder


@_logs.logged(desc_detail="{s3_prefix}->{local_folder}")
def download_s3_folder(s3_prefix, local_folder):
    for s3_file in list_s3_files(s3_prefix):
        target_file = _os.path.join(local_folder, _os.path.basename(s3_file))
        download_s3_file(s3_file, target_file)
    return local_folder


def download_git_repo(repo_url, git_ref, target_dir):
    _jobs.run_command(f"git clone https://{repo_url} .", cwd=target_dir)
    if git_ref != "master":
        _jobs.run_command(f"git fetch", cwd=target_dir)
        _jobs.run_command(f"git checkout {git_ref}", cwd=target_dir)


def download_git_folder(git_path, local_folder):
    repo_url, git_ref, code_path = parse_git_path(git_path)
    temp_folder = get_temp_dir()
    download_git_repo(repo_url, git_ref, temp_folder)
    copy_folder(_os.path.join(temp_folder, code_path), local_folder)
    return local_folder


# Function wrappers for cloud IO
def s3write_using(func, *args, **kwargs):
    """ Send any function's output to s3 """
    newargs = []
    newkwargs = kwargs.copy()
    temp_path_map = {}
    for arg in args:
        if isinstance(arg, str) and arg.startswith("s3"):
            tmppath = _tempfile.NamedTemporaryFile(delete=False).name
            temp_path_map[tmppath] = str(arg.replace("s3a:", "s3:"))
            newargs.append(tmppath)
        else:
            newargs.append(arg)
    for k, v in kwargs.items():
        if isinstance(v, str) and v.startswith("s3"):
            tmppath = _tempfile.NamedTemporaryFile(delete=False).name
            temp_path_map[tmppath] = str(v.replace("s3a:", "s3:"))
            newkwargs[k] = tmppath
    func(*newargs, **newkwargs)
    if temp_path_map:
        s3 = _boto3.client("s3")
        for local_path, s3_path in temp_path_map.items():
            if _SAFE_PATHS and not any([s3_path in safepath for safepath in _SAFE_PATHS]):
                raise RuntimeError(
                    f"Path '{s3_path}' cannot be written to because it is not in the "
                    f"designated safe output paths: '{', '.join(_SAFE_PATHS)}'"
                )
            _LOGGER.info(f"Uploading file to S3: {s3_path}")
            upload_s3_file(local_path, s3_path)
            _os.remove(local_path)
        _LOGGER.info(f"S3 upload(s) complete!")


def s3read_using(func, *args, **kwargs):
    """ Send any function's output to s3 """
    newargs = []
    newkwargs = kwargs.copy()
    tmpfolder = _tempfile.gettempdir()
    temp_path_map = {}
    for arg in args:
        if isinstance(arg, str) and arg.startswith("s3"):
            tmppath = _os.path.join(tmpfolder, _os.path.basename(arg))
            temp_path_map[tmppath] = str(arg.replace("s3a:", "s3:"))
            newargs.append(tmppath)
        else:
            newargs.append(arg)
    for k, v in kwargs.items():
        if isinstance(v, str) and v.startswith("s3"):
            tmppath = _os.path.join(tmpfolder, _os.path.basename(arg))
            temp_path_map[tmppath] = str(v.replace("s3a:", "s3:"))
            newkwargs[k] = tmppath
    if temp_path_map:
        for local_path, s3_path in temp_path_map.items():
            _LOGGER.info(f"Download file from S3: {s3_path}")
            download_s3_file(s3_path, local_path)
        _LOGGER.info(f"S3 download(s) complete!")
    _LOGGER.debug(
        "s3read_using() is running function with the following args:\n"
        f"{str(newargs)}\n{str(newkwargs)}"
    )
    result = func(*newargs, **newkwargs)
    if temp_path_map:
        for local_path, _ in temp_path_map.items():
            _LOGGER.info(f"Deleting temporary local file: {local_path}")
            _os.remove(local_path)
    return result


def _get_aws_settings_from_file(filepath, profile="default"):
    config = {}
    profle_match = False
    for line in get_text_file_contents(filepath).splitlines():
        if line.startswith(f"[{profile}]"):
            profile_match = True
            continue
        elif line.startswith("["):
            profile_match = False
            continue
        if profile_match and line.rstrip():
            line = line.split("#")[0].strip()
            if "=" in line:
                k, v = (s.strip() for s in line.split("="))
                config[k.upper()] = v
    return config


def load_aws_creds(update_env_vars=True):
    """ Load AWS creds. Returns a 2-item duple or None. """
    key, secret = None, None
    settings = {}
    default_creds_file = _os.environ.get(
        "AWS_SHARED_CREDENTIALS_FILE", _os.path.expanduser("~/.aws/credentials")
    )
    default_config_file = _os.environ.get(
        "AWS_CONFIG_FILE", _os.path.expanduser("~/.aws/config")
    )
    if file_exists(default_creds_file):
        _LOGGER.info(f"Parsing AWS credentials file '{default_creds_file}'...")
        settings.update(_get_aws_settings_from_file(default_creds_file))
    if file_exists(default_config_file):
        _LOGGER.info(f"Parsing AWS config file '{default_config_file}'...")
        settings.update(_get_aws_settings_from_file(default_config_file))
    if "AWS_ACCESS_KEY_ID" in _os.environ and "AWS_SECRET_ACCESS_KEY" in _os.environ:
        _LOGGER.info("Found env vars: 'AWS_ACCESS_KEY_ID' and 'AWS_SECRET_ACCESS_KEY'")
        settings["AWS_ACCESS_KEY_ID"] = _os.environ["AWS_ACCESS_KEY_ID"]
        settings["AWS_SECRET_ACCESS_KEY"] = _os.environ["AWS_SECRET_ACCESS_KEY"]
    if "AWS_ACCESS_KEY_ID" not in settings or "AWS_SECRET_ACCESS_KEY" not in settings:
        raise RuntimeError(f"Could not find AWS creds in file or env variables.")
    key = settings["AWS_ACCESS_KEY_ID"]
    secret = settings["AWS_SECRET_ACCESS_KEY"]
    if update_env_vars:
        _os.environ["AWS_ACCESS_KEY_ID"] = key
        _os.environ["AWS_SECRET_ACCESS_KEY"] = secret
    return key, secret


def main():
    _fire.Fire()


if __name__ == "__main__":
    main()



""" Standard sparkutils module """

import datetime
import hashlib
import json
import time
import tempfile
import os
from shutil import copyfile
import sys

import docker
import fire

from slalom.dataops.logs import get_logger, logged, logged_block
from slalom.dataops import jobs
from slalom.dataops import io

MAX_ECS_WAIT = 12 * 60 * 60  # max 12 hours wait

docker_client = docker.from_env()
logging = get_logger("slalom.dataops.dockerutils")


def build(dockerfile_path, tag_as, addl_args=None):
    """ Build an image. 'tag_as' can be a string or list of strings """
    folder_path = os.path.dirname(dockerfile_path)
    addl_args = addl_args or ""
    tag_as = _to_list(tag_as)
    if tag_as:
        tags = " ".join([f"-t {t}" for t in tag_as])
        cmd = f"docker build {addl_args} {tags} {folder_path} -f {dockerfile_path}"
    else:
        cmd = f"docker build {addl_args} {folder_path} -f {dockerfile_path}"
    jobs.run_command(cmd)


def _to_list(str_or_list):
    if str_or_list is None:
        return []
    elif isinstance(str_or_list, str):
        return str_or_list.split(",")
    else:
        return str_or_list


def tag(image_name: str, tag_as):
    """ Tag an image. 'tag_as' can be a string or list of strings """
    tag_as = _to_list(tag_as)
    for tag in tag_as:
        jobs.run_command(f"docker tag {image_name} {tag}")


@logged("pushing image '{image_name}'")
def push(image_name):
    # docker_client.images.push(image_name)
    cmd = f"docker push {image_name}"
    jobs.run_command(cmd)


def smart_split(dockerfile_path: str, tag_as, addl_args=None):
    tag_as = _to_list(tag_as)
    if tag_as:
        interim_image_name = tag_as[0].split(":")[0]
    else:
        interim_image_name = "untitled_image"
    (image_core, dockerfile_core), (image_derived, dockerfile_derived) = _smart_split(
        dockerfile_path, interim_image_name, addl_args=addl_args
    )
    dockerfile_path_core = os.path.realpath(f"{dockerfile_path}.core")
    dockerfile_path_derived = os.path.realpath(f"{dockerfile_path}.quick")
    io.create_text_file(filepath=dockerfile_path_core, contents=dockerfile_core)
    if dockerfile_derived:
        io.create_text_file(filepath=dockerfile_path_derived, contents=dockerfile_derived)
    else:
        io.delete_file(dockerfile_path_derived, ignore_missing=True)
        dockerfile_path_derived = None
    return image_core, dockerfile_path_core, image_derived, dockerfile_path_derived


@logged("smartly building '{dockerfile_path}' as {tag_as or '(none)'}")
def smart_build(
    dockerfile_path: str,
    tag_as=None,
    push_core=True,
    push_final=False,
    with_login=False,
    addl_args=None,
    ignore_caches=False,
):
    """
    Builds the dockerfile if needed but pulls it from the remote if possible.
    """
    if bool(with_login):
        login()
    tag_as = _to_list(tag_as)
    result = smart_split(dockerfile_path, tag_as, addl_args=addl_args)
    image_core, dockerfile_path_core, image_derived, dockerfile_path_derived = result
    if not ignore_caches:
        if dockerfile_path_derived is None and exists_remotely(image_core):
            logging.info(
                "Image with matching hash already exists "
                "and no host files are referenced in Dockerfile."
                f"Attempting to retag existing image '{image_core}' as '{tag_as}'..."
            )
            return remote_retag(
                image_name=image_core.split(":")[0],
                existing_tag=image_core.split(":")[1],
                tag_as=tag_as,
            )
        pull(image_core, skip_if_exists=True, silent=True)
    if ignore_caches or not exists_locally(image_core):
        with logged_block(f"building interim (core) image as '{image_core}'"):
            build(dockerfile_path_core, image_core, addl_args=addl_args)
    if push_core:
        if ignore_caches or not exists_remotely(image_core):
            with logged_block(f"pushing interim (core) image '{image_derived}'"):
                push(image_core)
        else:
            logging.info(f"Already exists. Skipping push of image '{image_derived}'")
    with logged_block(f"building '{dockerfile_path_derived}' as '{image_derived}'"):
        if dockerfile_path_derived:
            build(dockerfile_path_derived, image_derived, addl_args=addl_args)
        else:
            tag(image_core, image_derived)
    if tag_as:
        tag(image_derived, tag_as)
        if push_final:
            for image_name in tag_as:
                push(image_name)


@logged("pulling image {image_name}")
def pull(image_name, skip_if_exists=False, silent=False):
    if skip_if_exists and exists_locally(image_name):
        logging.info(f"Skipping image pull. Already exists locally: {image_name}")
        return image_name
    else:
        try:
            jobs.run_command(f"docker pull {image_name}", raise_error=True)
        except Exception as ex:
            logging.info(f"Failed to pull image: {image_name}\n{ex}")
            if silent:
                return False
            else:
                raise ex
        if not exists_locally(image_name):
            logging.warning("Pull was successful in API but could not be confirmed")
        return image_name


def exists_locally(image_name):
    try:
        image = docker_client.images.get(image_name)
        return True
    except docker.errors.ImageNotFound as ex:
        return False


def exists_remotely(image_name):
    try:
        image = docker_client.images.get_registry_data(image_name)
        if image:
            return True
        else:
            return False
    except docker.errors.ImageNotFound as ex:
        return False
    except Exception as ex:
        logging.exception(
            f"Failure when checking if image exists remotely '{image_name}'"
        )
        return None


def _smart_split(dockerfile_path, image_name, addl_args=None):
    """ 
    Returns list of tuples: [
        (partial_image_name, partial_dockerfile_text)
        (derived_image_name, derived_dockerfile_text)
    ]
    Create two dockerfiles from a single file.
    1. The first 'core' image will contain all statements until the first COPY or ADD. 
    2. The second 'derived' image will pull from 'core' and complete the build using 
    local files or artifacts required by ADD or COPY commands.
    """
    orig_text = io.get_text_file_contents(dockerfile_path)
    addl_args = addl_args or ""
    core_dockerfile = ""
    derived_dockerfile = ""
    requires_context = False  # Whether we need file context to determine output
    for line in orig_text.split("\n"):
        if any([line.startswith("COPY"), line.startswith("ADD")]):
            requires_context = True
        if not requires_context:
            core_dockerfile += line + "\n"
        else:
            derived_dockerfile += line + "\n"
    core_md5 = hashlib.md5((addl_args + core_dockerfile).encode("utf-8")).hexdigest()
    full_md5 = hashlib.md5((addl_args + orig_text).encode("utf-8")).hexdigest()
    core_image_name = f"{image_name}:core-md5-{core_md5}"
    derived_image_name = f"{image_name}:md5-{full_md5}"

    core_dockerfile = (
        f"# NO NOT EDIT - file is generated automatically from `Dockerfile`\n\n"
        f"# Dockerfile.core - will be created and pushed as:\n"
        f"# \t{core_image_name}\n\n{core_dockerfile}"
    )
    if derived_dockerfile:
        derived_dockerfile = (
            f"# NO NOT EDIT - file is generated automatically from `Dockerfile`\n\n"
            f"FROM {core_image_name}\n\n{derived_dockerfile}"
        )
    else:
        derived_dockerfile = None  # No additional work to do.
    return [(core_image_name, core_dockerfile), (derived_image_name, derived_dockerfile)]


def ecs_login(region):
    logging.info("Logging into ECS...")
    try:
        _, ecs_login_cmd = jobs.run_command(
            f"aws ecr get-login --region {region} --no-include-email", echo=False
        )
        _, _ = jobs.run_command(ecs_login_cmd, hide=True)
    except Exception as ex:
        raise RuntimeError(f"ECS login failed. {ex}")


def login(raise_error=False):
    usr = os.environ.get("DOCKER_USERNAME", "")
    pwd = os.environ.get("DOCKER_PASSWORD", "")
    registry = os.environ.get("DOCKER_REGISTRY", "") or "index.docker.io"
    if not (usr and pwd):
        error_msg = (
            "Could not login to docker registry."
            "Missing env variable DOCKER_USERNAME or DOCKER_PASSWORD"
        )
        if raise_error:
            raise RuntimeError(error_msg)
        else:
            logging.warning(error_msg)
            return False
    logging.info(f"Logging into docker registry '{registry}' as user '{usr}'...")
    try:
        jobs.run_command(
            f"docker login {registry} --username {usr} --password {pwd}", hide=True
        )
        if registry == "index.docker.io":
            jobs.run_command(f"docker login --username {usr} --password {pwd}", hide=True)
    except Exception as ex:
        if raise_error:
            raise RuntimeError(f"Docker login failed. {ex}")
        else:
            logging.warning(f"Docker login failed. {ex}")


@logged("applying tag '{tag_as}' to remote ECS image '{image_name}:{existing_tag}'")
def ecs_retag(image_name, existing_tag, tag_as):
    tag_as = _to_list(tag_as)
    if "amazonaws.com/" in image_name:
        image_name = image_name.split("amazonaws.com/")[1]
    get_manifest_cmd = (
        f"aws ecr batch-get-image"
        f" --repository-name {image_name} --image-ids imageTag={existing_tag}"
        f" --query 'images[].imageManifest' --output text"
    )
    _, manifest = jobs.run_command(get_manifest_cmd, echo=False)
    for new_tag in tag_as:
        if "amazonaws.com/" in new_tag:
            new_tag = new_tag.split("amazonaws.com/")[1]
        if ":" in new_tag:
            if image_name != new_tag.split(":")[0]:
                raise RuntimeError(
                    f"Image names do not match: '{image_name}', '{new_tag.split(':')[0]}'"
                )
            new_tag = new_tag.split(":")[1]
        put_image_cmd = [
            "aws",
            "ecr",
            "put-image",
            "--repository-name",
            image_name,
            "--image-tag",
            new_tag,
            "--image-manifest",
            manifest,
        ]
        return_code, output_text = jobs.run_command(
            put_image_cmd, shell=False, echo=False, hide=True, raise_error=False
        )
        if return_code != 0 and "ImageAlreadyExistsException" in output_text:
            logging.info("Image already exists. No tagging changes were made.")
        elif return_code != 0:
            raise RuntimeError(f"Could not retag the specified image.\n{output_text}")


@logged("applying tag '{tag_as}' to remote image '{image_name}:{existing_tag}'")
def remote_retag(image_name, existing_tag, tag_as, with_login=False):
    tag_as = _to_list(tag_as)
    if bool(with_login):
        login()
    if "amazonaws.com/" in image_name:
        return ecs_retag(image_name, existing_tag, tag_as)
    existing_fullname = f"{image_name}:{existing_tag}"
    pull(existing_fullname)
    for new_tag in tag_as:
        if ":" in new_tag:
            new_fullname = new_tag
        else:
            new_fullname = f"{image_name}:{new_tag}"
        tag(existing_fullname, new_fullname)
        push(new_fullname)


def ecs_submit(
    task_name: str,
    cluster: str,
    region: str,
    container_name: str = None,
    cmd_override: dict = None,
    env_overrides: dict = None,
    use_fargate: str = False,
    wait_for_start=True,
    wait_for_stop=False,
    max_wait=None,
    yyyymmdd=None,
):
    cmd = (
        f"aws ecs run-task"
        f" --task-definition {task_name}"
        f" --cluster {cluster}"
        f" --region {region}"
    )
    if use_fargate:
        cmd += f" --launch-type FARGATE"
    else:
        cmd += f" --launch-type EC2"
    if env_overrides and isinstance(env_overrides, str):
        env_overrides = {
            x.split("=")[0]: x.split("=")[1] for x in env_overrides.split(",")
        }
    if yyyymmdd and yyyymmdd != "0":
        if str(yyyymmdd).lower() == "today":
            yyyymmdd = datetime.today().strftime("%Y%m%d")
        env_overrides = env_overrides or {}
        env_overrides["YYYYMMDD"] = yyyymmdd
    if env_overrides or cmd_override:
        if not container_name:
            raise ValueError(
                "container_name is required if "
                "cmd_override or env_overrides are specified"
            )
        env_override_str = ""
        cmd_override_str = ""
        if env_overrides:
            env_override_str = (
                ',"environment":['
                + ",".join(
                    [
                        "{" + f'"name":"{k}","value":"{v}"' + "}"
                        for k, v in env_overrides.items()
                    ]
                )
                + "]"
            )
        if cmd_override:
            cmd_override_str = f", 'command': ['{cmd_override}']"
        overrides = (
            ' --overrides \'{"containerOverrides":'
            f'[{{"name":"{container_name}"'
            f"{cmd_override_str}{env_override_str}"
            "}]}'"
        )
        cmd += overrides
    return_code, output_text = jobs.run_command(cmd, raise_error=False, echo=False)
    if return_code != 0:
        raise RuntimeError(f"Could not start task: {output_text}")
    jsonobj = json.loads(output_text)
    if len(jsonobj.get("tasks", [])) == 0 or len(jsonobj.get("failures", [])) > 0:
        raise RuntimeError(
            f"Could not start task ({jsonobj.get('failures', '')})\n{output_text}"
        )
    task_arn = jsonobj["tasks"][0]["taskArn"]
    logging.info(f"ECS task status: {get_ecs_task_detail_url(region, task_arn, cluster)}")
    logging.info(f"ECS task logs:   {get_ecs_log_url(region, task_arn)}")
    if wait_for_start:
        ecs_wait_for_start(task_arn=task_arn, cluster=cluster, region=region)
    if wait_for_stop:
        ecs_wait_for_stop(task_arn=task_arn, cluster=cluster, region=region)
    if not wait_for_start and not wait_for_stop:
        logging.debug(f"ECS submit result: {output_text}")
    return task_arn


def get_ecs_log_url(
    region,
    task_arn,
    container_name="PTB-Container",
    log_group="PTB-AWSLogs20190822233355860300000001",
):
    task_id = task_arn.split("/")[-1]
    return (
        f"https://{region}.console.aws.amazon.com/cloudwatch/home?"
        f"region={region}#logEventViewer:group={log_group};"
        f"stream=container-log/{container_name}/{task_id}"
    )


def get_ecs_task_detail_url(region, task_arn, cluster_name):
    task_id = task_arn.split("/")[-1]
    return (
        f"https://{region}.console.aws.amazon.com/ecs/home?"
        f"region={region}#/clusters/{cluster_name}/tasks/{task_id}/details"
    )


def ecs_wait_for_start(task_arn, cluster, region, timeout=1200, raise_error=True):
    return _ecs_wait_for(
        "running",
        task_arn,
        cluster,
        region,
        timeout=timeout,
        heartbeat_interval=15,
        raise_error=raise_error,
    )


def ecs_wait_for_stop(task_arn, cluster, region, timeout=1200, raise_error=True):
    return _ecs_wait_for(
        "stopped",
        task_arn,
        cluster,
        region,
        timeout=timeout,
        heartbeat_interval=2 * 60,
        raise_error=raise_error,
    )


@logged(
    "waiting for ECS status '{wait_for}'",
    success_detail=lambda: get_ecs_log_url("{region}", "{task_arn}"),
)
def _ecs_wait_for(
    wait_for,
    task_arn,
    cluster,
    region,
    timeout=1200,
    heartbeat_interval=None,
    raise_error=True,
):
    task_id = task_arn.split("/")[-1]
    wait_cmd = f"aws ecs wait tasks-{wait_for} --cluster {cluster} --tasks {task_arn}"
    desc_cmd = f"aws ecs describe-tasks --cluster {cluster} --tasks {task_arn}"

    with logged_block(
        f"waiting for ECS job to reach '{wait_for}' status",
        heartbeat_interval=heartbeat_interval,
    ):
        timeout_time = time.time() + (timeout or MAX_ECS_WAIT)
        return_code, output_text = jobs.run_command(wait_cmd, raise_error=False)
        while return_code == 255 and time.time() < timeout_time:
            logging.info("aws cli timeout expired. Retrying...")
            return_code, output_text = jobs.run_command(wait_cmd, raise_error=True)
        if return_code != 0:
            raise RuntimeError(
                f"ECS wait command failed or timed out (return={return_code}).\n"
                f"{output_text}"
            )
    return_code, output_text = jobs.run_command(desc_cmd, raise_error=False)
    if return_code != 0:
        raise RuntimeError(f"ECS task describe failed.\n{output_text}")

    jsonobj = json.loads(output_text)
    if len(jsonobj.get("tasks", [])) == 0 or len(jsonobj.get("failures", [])) > 0:
        RuntimeError(f"Could not start task ({jsonobj.get('failures', '')})")
    task_arn = jsonobj["tasks"][0]["taskArn"]
    logging.info(f"ECS task status: {get_ecs_task_detail_url(region, task_arn, cluster)}")
    logging.info(f"ECS task logs:   {get_ecs_log_url(region, task_arn)}")
    return task_arn


def main():
    fire.Fire()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import os
import platform
import sys

import fire

from slalom.dataops.logs import get_logger, logged, logged_block
from slalom.dataops import jobs, io

if os.name == "nt":
    import ctypes
else:
    ctypes = None


CACHED_INSTALL_LIST: str = None
DEBUG = False
logging = get_logger("slalom.dataops.env", debug=DEBUG)

CHOCO_INSTALL_CMD = """
@"%SystemRoot%\System32\WindowsPowerShell\v1.0\powershell.exe" -NoProfile -InputFormat None -ExecutionPolicy Bypass -Command "iex ((New-Object System.Net.WebClient).DownloadString('https://chocolatey.org/install.ps1'))" && SET "PATH=%PATH%;%ALLUSERSPROFILE%\chocolatey\bin"
""".strip()
WINDOWS_INSTALL_LIST = {
    "chocolatey": CHOCO_INSTALL_CMD,
    "chocolateygui": None,
    "sudo": None,
    "git": 'choco install -y git.install --params "/GitOnlyOnPath /SChannel /NoAutoCrlf /WindowsTerminal"',
    "docker": "docker-desktop",
    "python": "python3",
    "terraform": None,
    "vscode": None,
}
LINUX_INSTALL_LIST = {"docker": "docker.io", "docker-compose": "docker-compose"}
MAC_INSTALL_LIST = {"docker": "brew cask install docker && open /Applications/Docker.app"}


def is_windows() -> bool:
    return platform.system() == "Windows"


def is_mac() -> bool:
    return platform.system() == "Darwin"


def is_linux() -> bool:
    return platform.system() == "Linux"


def status():
    check_installs(install_if_missing=False)


def _to_list(str_or_list):
    if str_or_list is None:
        return []
    elif isinstance(str_or_list, str):
        return str_or_list.split(",")
    else:
        return str_or_list


def is_admin():
    if is_linux() or is_mac():
        return os.geteuid() == 0
    elif is_windows():
        try:
            return ctypes.windll.shell32.IsUserAnAdmin()
        except:
            return False
    else:
        return None


def rerun_as_admin(prompt=True):
    if is_admin():
        return
    else:  # Re-run the program with admin rights
        args = " ".join(sys.argv[1:])
        response = "y"
        if prompt:
            response = input(
                f"While executing '{sys.executable} {__file__} {args}' we detected "
                f"that you do not have admin/root permissions.\n"
                f"Would you like to rerun this program as admin? [y/n]"
            ).lower()
        if response in ["y", "yes"]:
            return run_as_admin(
                wait=True, cmd=[sys.executable] + [__file__] + sys.argv[1:]
            )
        elif response in ["n", "no"]:
            return None
        else:
            raise RuntimeError("Invalid response.")


@logged("running {' '.join(cmd or [])}")
def run_as_admin(cmd: list = None, wait=True):

    if os.name != "nt":
        raise NotImplementedError("This function is only implemented on Windows.")

    import win32api, win32con, win32event, win32process
    from win32com.shell.shell import ShellExecuteEx
    from win32com.shell import shellcon
    import types

    if cmd is None:
        cmd = [sys.executable] + sys.argv
    process = '"%s"' % (cmd[0],)
    params = " ".join(['"%s"' % (x,) for x in cmd[1:]])
    cmdDir = ""
    showCmd = win32con.SW_SHOWNORMAL
    procInfo = ShellExecuteEx(
        nShow=showCmd,
        fMask=shellcon.SEE_MASK_NOCLOSEPROCESS,
        lpVerb="runas",
        lpFile=process,
        lpParameters=params,
    )
    if wait:
        procHandle = procInfo["hProcess"]
        obj = win32event.WaitForSingleObject(procHandle, win32event.INFINITE)
        rc = win32process.GetExitCodeProcess(procHandle)
        print(f"Return code: {rc}")
    else:
        rc = None
    return rc


def check_installs(install_list=None, install_if_missing: bool = None):
    if is_windows():
        ref_list = WINDOWS_INSTALL_LIST
    elif is_linux():
        ref_list = LINUX_INSTALL_LIST
    elif is_mac():
        ref_list = MAC_INSTALL_LIST
    else:
        raise RuntimeError("Could not detect OS type.")
    install_list = {
        program: ref_list.get(program.lower(), None)
        for program in (_to_list(install_list) or ref_list.keys())
    }
    missing = {}
    for name, install_cmd in install_list.items():
        installed = check_install(name, install_if_missing, install_cmd)
        logging.info(f"{name}: {'OK' if installed else 'Not installed'}")
        if not installed:
            missing[name] = install_cmd
    if install_if_missing is None and missing:  # prompt if not specified
        if input(
            f"The following software components appear to be missing: {','.join(missing.keys())}\n"
            f"Would you like to install them? [y/n] "
        ).lower() in ["y", "yes"]:
            for program, install_cmd in missing.items():
                install(program, install_cmd)


def check_install(
    program_name: str, install_if_missing: bool = None, install_cmd: str = None
) -> bool:
    installed = False
    installed_programs = get_installed_programs()
    installed = program_name.lower() in installed_programs.keys()
    if not installed:
        for test_cmd in [f"{program_name} --version", f"which {program_name}"]:
            return_code, output = jobs.run_command(test_cmd, raise_error=False)
            if return_code == 0 and len(output) > 1:
                installed = True
                break
    if install_if_missing and not installed:
        install(program_name, install_cmd)
        installed = True
    return installed


def _default_install_cmd(program_name):
    if is_windows():
        return f"choco install -y {program_name}"
    elif is_linux():
        return f"apt-get install -y {program_name}"
    elif is_mac():
        return f"choco install {program_name}"
    else:
        raise RuntimeError("Could not detect OS type.")


def get_installed_programs():
    global CACHED_INSTALL_LIST

    if CACHED_INSTALL_LIST:
        return CACHED_INSTALL_LIST
    if is_windows():
        return_code, output = jobs.run_command("choco list --local", raise_error=False)
        if return_code == 0:
            CACHED_INSTALL_LIST = {
                x.split(" ")[0].lower(): x.split(" ")[1]
                for x in output.split("\n")
                if len(x.split(" ")) == 2
            }
        else:
            CACHED_INSTALL_LIST = {}
    elif is_linux():
        CACHED_INSTALL_LIST = {}
    elif is_mac():
        CACHED_INSTALL_LIST = {}
    return CACHED_INSTALL_LIST


@logged("installing {program_name}")
def install(program_name, install_cmd):
    if not install_cmd:
        install_cmd = _default_install_cmd(program_name=program_name)
    elif " " not in install_cmd:
        install_cmd = _default_install_cmd(program_name=install_cmd)
    if not is_admin():
        return_code = run_as_admin(cmd=install_cmd.split(" "), prompt=True)
    else:
        return_code, output = jobs.run_command(install_cmd)
    return return_code == 0


if __name__ == "__main__":
    fire.Fire()



#!/usr/bin/env python3
from joblib import Parallel, delayed
import os
import sys
from pathlib import Path

import fire
from tqdm import tqdm

code_file = os.path.realpath(__file__)
repo_dir = os.path.dirname(os.path.dirname(os.path.dirname(code_file)))
sys.path.append(os.path.join(repo_dir, "src"))
sys.path.append(os.path.dirname(code_file))

from slalom.dataops.logs import get_logger, logged, logged_block
from slalom.dataops import jobs, io


DEBUG = False
logging = get_logger("slalom.dataops.infra", debug=DEBUG)


def _update_var_output(output_var):
    return_code, output = jobs.run_command(f"terraform output {output_var}", echo=False)
    io.create_text_file(os.path.join("outputs", output_var), contents=output)
    return True


@logged("updating output files")
def update_var_outputs(infra_dir, output_vars=[]):
    outputs_dir = os.path.join(infra_dir, "outputs")
    io.create_folder(outputs_dir)
    for oldfile in io.list_files(outputs_dir):
        io.delete_file(oldfile)
    results = Parallel(n_jobs=40, verbose=2)(
        delayed(_update_var_output)(outvar)
        for outvar in tqdm(output_vars, "Saving output to infra/outputs")
    )


def install(*args, infra_dir="./infra", deploy=False, git_ref="master"):
    """
    Usage example:
    ```
    s-infra install catalog:aws-prereqs --infra_dir=infra/prereqs --deploy=True
    s-infra install samples:aws --infra_dir=infra --deploy=True
    ```
    Which is identical to:
    ```
    s-infra install catalog:aws-prereqs --infra_dir=infra/prereqs
    s-infra init+apply --infra_dir=infra/prereqs
    s-infra install samples:aws --infra_dir=infra
    s-infra init+apply --infra_dir=infra
    ```
    """
    io.create_folder(infra_dir)
    for arg in args:
        with logged_block(f"installing terraform modules from '{arg}'"):
            infra_type, infra_name = arg.split(":")
            if infra_type not in ["catalog", "samples"]:
                raise ValueError(
                    f"Expected infra_type to be one of: 'catalog', 'samples'. Received type: {infra_type}"
                )
            io.download_folder(
                remote_folder=f"git://github.com/slalom-ggp/dataops-infra#{git_ref}//{infra_type}/{infra_name}",
                local_folder=infra_dir,
            )
    lf = "\n"
    logging.info(f"List of installed modules:\n{lf.join(io.ls(infra_dir))}")
    init(infra_dir=infra_dir)
    if deploy:
        apply(infra_dir=infra_dir)


@logged("initializing terraform project at '{infra_dir}'")
def init(infra_dir: str = "./infra/"):
    infra_dir = os.path.realpath(infra_dir)
    jobs.run_command("terraform init", working_dir=infra_dir)


@logged("applying terraform changes")
def apply(infra_dir: str = "./infra/", save_output: bool = False, prompt: bool = False):
    infra_dir = os.path.realpath(infra_dir)
    jobs.run_command(
        f"terraform apply {'' if prompt else '-auto-approve'}", working_dir=infra_dir
    )
    if save_output:
        update_var_outputs(infra_dir=infra_dir)


def init_and_apply(infra_dir: str = "./infra/", save_output: bool = False):
    infra_dir = os.path.realpath(infra_dir)
    init(infra_dir=infra_dir)
    apply(infra_dir=infra_dir, save_output=save_output, prompt=False)


def change_upstream_source(
    dir_to_update=".",
    git_repo="https://github.com/slalom-ggp/dataops-infra",
    branch="master",
    relative_path="../../dataops-infra",
    to_relative=False,
    to_git=False,
    dry_run=False,
):
    """Change Terraform source"""
    if to_relative and to_git or not (to_relative or to_git):
        raise ValueError("Must specify `--to_git` or `--to_relative`, but not both.")
    for tf_file in io.list_files(dir_to_update):
        if tf_file.endswith(".tf"):
            # print(tf_file)
            new_lines = []
            for line in io.get_text_file_contents(tf_file).splitlines():
                new_line = line
                if line.lstrip().startswith("source "):
                    current_path = line.lstrip().split('"')[1]
                    start_pos = max(
                        [current_path.find("catalog/"), current_path.find("components/")]
                    )
                    if start_pos > 0:
                        module_path = current_path[start_pos:].split("?ref=")[0]
                        if to_relative:
                            local_patten = "{relative_path}/{path}"
                            new_path = local_patten.format(
                                relative_path=relative_path, path=module_path
                            )
                        elif to_git:
                            git_pattern = "git::{git_repo}//{path}?ref={branch}"
                            new_path = git_pattern.format(
                                git_repo=git_repo, path=module_path, branch=branch
                            )
                        print(f"{current_path} \n\t\t\t>> {new_path}")
                        new_line = f'  source = "{new_path}"'
                new_lines.append(new_line)
            new_file_text = "\n".join(new_lines)
            if dry_run:
                print(f"\n\n------------\n-- {tf_file}\n------------")
                print(new_file_text)
            else:
                io.create_text_file(tf_file, new_file_text)
    if not dry_run:
        jobs.run_command("terraform fmt -recursive", dir_to_update)


def main():
    fire.Fire(
        {
            "install": install,
            "init": init,
            "apply": apply,
            "init+apply": init_and_apply,
            "deploy": init_and_apply,
            "change_upstream_source": change_upstream_source,
        }
    )


if __name__ == "__main__":
    main()



from setuptools import setup
import os
from pathlib import Path

detected_version = None
version_filepath = "VERSION"

if "VERSION" in os.environ:
    detected_version = os.environ["VERSION"]
    if "/" in detected_version:
        detected_version = detected_version.split("/")[-1]
if not detected_version and os.path.exists(version_filepath):
    detected_version = Path(version_filepath).read_text()
    if len(detected_version.split(".")) <= 3:
        if "BUILD_NUMBER" in os.environ:
            detected_version = f"{detected_version}.{os.environ['BUILD_NUMBER']}"
if not detected_version:
    raise RuntimeError("Error. Could not detect version.")
if os.environ.get("BRANCH_NAME", "unknown") not in ["master", "refs/heads/master"]:
    detected_version = f"{detected_version}.dev0"

detected_version = detected_version.lstrip("v")
print(f"Detected version: {detected_version}")
Path(version_filepath).write_text(f"v{detected_version}")

setup(
    name="slalom.dataops",
    packages=["slalom.dataops"],
    version=detected_version,
    license="MIT",
    description="Slalom GGP libary for DataOps automation",
    author="AJ Steers",
    author_email="aj.steers@slalom.com",
    url="https://bitbucket.org/slalom-consulting/dataops-tools/",
    download_url="https://github.com/slalom-ggp/dataops-tools/archive/v_0.1.tar.gz",
    keywords=["DATAOPS", "SLALOM", "DATA", "AUTOMATION", "CI/CD", "DEVOPS"],
    package_data={"": [version_filepath]},
    entry_points={
        "console_scripts": [
            # Register CLI commands:
            "s-docker = slalom.dataops.dockerutils:main",
            "s-infra = slalom.dataops.infra:main",
            "s-io = slalom.dataops.io:main",
            "s-spark = slalom.dataops.sparkutils:main",
            "s-tap = slalom.dataops.taputils:main",
        ]
    },
    include_package_data=True,
    install_requires=[
        "docker",
        "fire",
        "joblib",
        "junit-xml",
        "psutil",
        "pyyaml",
        "tqdm",
        "xmlrunner",
    ],
    extras_require={
        "AWS": ["awscli", "s3fs"],
        "Azure": ["azure"],
        "Pandas": ["pandas"],
        "Spark": ["pyspark"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
)
